import sys

import numpy as np

from deep_utils.utils.box_utils.boxes import Box, Point
from deep_utils.utils.lib_utils.download_utils import download_decorator
from deep_utils.utils.lib_utils.lib_decorators import (
    expand_input,
    get_elapsed_time,
    get_from_config,
    rgb2bgr,
)
from deep_utils.vision.face_detection.main.main_face_detection import FaceDetector

from .config import Config


class RetinaFaceTorchFaceDetector(FaceDetector):
    def __init__(self, **kwargs):
        import torch

        from .src.layers.functions.prior_box import PriorBox
        from .src.load_model import load_model
        from .src.retinaface import RetinaFace
        from .src.utils.box_utils import decode, decode_landm
        from .src.utils.nms.py_cpu_nms import py_cpu_nms

        self.RetinaFace = RetinaFace
        self.load_model_weights = load_model
        self.torch = torch
        self.PriorBox = PriorBox
        self.py_cpu_nms = py_cpu_nms
        self.decode = decode
        self.decode_landm = decode_landm
        self.resize = 1
        self.pretrained_mobilenet = None
        network = Config.network.lower()

        if network == "resnet50":
            download_variables = ("resnet50",)
        elif network == "mobilenet":
            download_variables = ("mobilenet", "mobilenetV1X")
        else:
            raise ValueError(
                "The model is not supported\nsupported models: resnet50, mobilenet"
            )
        super().__init__(
            name=self.__class__.__name__,
            file_path=__file__,
            download_variables=download_variables,
            **kwargs
        )
        self.config: Config

    @download_decorator
    def load_model(self):
        cfg_resnet = Config.cfg_re50
        cfg_mobilenet = Config.cfg_mnet
        device = self.config.device
        network = str.lower(self.config.network)
        load_to_cpu = False if device == "cuda" else True
        if network == "resnet50":
            cfg, model_path = (cfg_resnet, self.config.resnet50)
            net = self.RetinaFace(cfg=cfg, phase="test")
            net = self.load_model_weights(
                net, model_path, load_to_cpu=load_to_cpu)
        elif network == "mobilenet":
            cfg, model_path = cfg_mobilenet, self.config.mobilenet
            net = self.RetinaFace(
                cfg=cfg, phase="test", pretrained_model=self.config.mobilenetV1X
            )
            net = self.load_model_weights(
                net, model_path, load_to_cpu=load_to_cpu)
        else:
            print("The network backbone is wrong!")
            sys.exit(1)
        self.config.cfg = cfg
        net.eval()
        net.to(device)
        self.config.priorbox = self.PriorBox(self.config.cfg)
        self.model = net

    @get_from_config
    @get_elapsed_time
    @rgb2bgr("bgr")
    @expand_input(3)
    def detect_faces(
        self,
        images,
        is_rgb,
        resize_size=None,
        img_scale_factor=None,
        img_mean=None,
        resize_mode=None,
        confidence=None,
        get_time=False,
        round_prec=4,
    ):
        device = self.config.device
        face_points = ["left_eye", "right_eye",
                       "nose", "mouth_left", "mouth_right"]
        images = np.float32(images)
        n_images = images.shape[0]
        images -= (104, 117, 123)
        images = images.transpose((0, 3, 1, 2))
        img = self.torch.FloatTensor(images).to(device)
        _, im_height, im_width = img[0].shape
        scale = self.torch.Tensor(
            [img[0].shape[2], img[0].shape[1], img[0].shape[2], img[0].shape[1]]
        )
        img = img.to(device)
        scale = scale.to(device)
        loc, conf, landms = self.model(img)

        priors = self.config.priorbox.forward(image_size=(im_height, im_width))
        priors = priors.to(device)
        prior_data = priors.data
        bboxes_, landmk_, scores_ = [], [], []
        for n_img in range(n_images):
            loc_data = loc.data[n_img]
            boxes = self.decode(loc_data, prior_data,
                                self.config.cfg["variance"])
            boxes = boxes * scale / self.resize
            boxes = boxes.cpu().numpy()
            scores = conf[n_img].data.cpu().numpy()[:, 1]

            landms = self.decode_landm(
                landms.data[n_img], prior_data, self.config.cfg["variance"]
            )
            scale1 = self.torch.Tensor(
                [
                    img.shape[3],
                    img.shape[2],
                    img.shape[3],
                    img.shape[2],
                    img.shape[3],
                    img.shape[2],
                    img.shape[3],
                    img.shape[2],
                    img.shape[3],
                    img.shape[2],
                ]
            )
            scale1 = scale1.to(device)
            landms = landms * scale1 / self.resize
            landms = landms.cpu().numpy()

            # ignore low scores
            inds = np.where(scores > confidence)[0]
            boxes = boxes[inds]
            landms = landms[inds]
            scores = scores[inds]

            # keep top-K before NMS
            order = scores.argsort()[::-1][: self.config.top_k]
            boxes = boxes[order]
            landms = landms[order]
            scores = scores[order]

            # do NMS
            bboxes = np.hstack((boxes, scores[:, np.newaxis])).astype(
                np.float32, copy=False
            )
            keep = self.py_cpu_nms(bboxes, self.config.nms_thresholds)
            bboxes = bboxes[keep, :]
            landms = landms[keep]

            # keep top-K faster NMS
            bboxes = bboxes[: self.config.keep_top_k, :]
            landms = landms[: self.config.keep_top_k, :]
            landmk_.append(landms)
            bboxes_.append(bboxes)
            scores_.append(scores)
        bboxes, landms, scores = np.array(
            bboxes_), np.array(landmk_), np.array(scores_)
        boxes_, confidences_, landmarks_ = [], [], []
        for num in range(len(bboxes)):
            box_tensor = bboxes[num, :]
            boxes = [[b[0], b[1], b[2], b[3]] for b in box_tensor]
            boxes = Box.box2box(
                boxes, in_source=Box.BoxSource.Torch, to_source=Box.BoxSource.Numpy
            )
            boxes_.append(boxes)
            confidences_.append(scores.round(round_prec))
            if len(landms) != 0:
                landmarks = [
                    [
                        Point.point2point(
                            (landms[num][j][2 * i], landms[num][j][2 * i + 1]),
                            in_source="Torch",
                            to_source="Numpy",
                        )
                        for i in range(5)
                    ]
                    for j in range(len(landms[num]))
                ]
            img_landmarks = []
            for i in range(len(landmarks)):
                face_dict = {}
                for points, face in zip(landmarks[i], face_points):
                    face_dict[face] = points
                img_landmarks.append(face_dict)
            landmarks_.append(img_landmarks)

        output = self.output_class(
            boxes=boxes_, confidences=confidences_, landmarks=landmarks_[00]
        )
        return output



import numpy as np
from deep_utils.utils.logging_utils.logging_utils import log_print
import glob
from PIL import Image


class GIFUtils:
    @staticmethod
    def make_gif_dir(frame_folder, output_path, extension=".jpg", logger=None, verbose=1):
        frames = [Image.open(image) for image in glob.glob(f"{frame_folder}/*.{extension}")]
        GIFUtils.make_gif(frames, output_path, logger=logger, verbose=verbose)

    @staticmethod
    def make_gif(frames, output_path, duration=None, logger=None, verbose=1):
        duration = duration if duration is not None else (len(frames) // 10)
        if isinstance(frames[0], np.ndarray):
            frames = [Image.fromarray(frame) for frame in frames]
        frame_one = frames[0]
        frame_one.save(output_path, format="GIF", append_images=frames,
                       save_all=True, duration=duration, loop=0)
        log_print(logger, f"Successfully saved gif to {output_path}", verbose=verbose)



def load_tf_opt(opt_name: str, lr: float, **args):
    """
    This module is used for loading tensorflow/keras optimizers
    Args:
        opt_name:
        lr:
        **args:

    Returns:

    """
    from tensorflow.keras.optimizers import SGD, Adadelta, Adam, RMSprop

    opt_dict = {"adam": Adam, "rmsprop": RMSprop,
                "sgd": SGD, "adadelta": Adadelta}
    opt = opt_dict.get(opt_name.lower(), None)
    if opt is None:
        raise ValueError(
            f"[ERROR] optimization: {opt_name} is not supported, supported optimization are {opt_dict.keys()}"
        )
    return opt(learning_rate=lr, **args)



import traceback
from os.path import join

import numpy as np


class TensorboardUtils:
    @staticmethod
    def save_image(df, path):
        import matplotlib.pyplot as plt
        from pandas.plotting import table

        ax = plt.subplot(111, frame_on=False)  # no visible frame
        ax.xaxis.set_visible(False)  # hide the x axis
        ax.yaxis.set_visible(False)  # hide the y axis
        table(ax, df)  # where df is your data frame
        plt.savefig(join(path, "csv_log.png"))

    @staticmethod
    def get_docx(df, doc_path):
        import docx

        # open an existing document
        doc = docx.Document(None)

        # add a table to the end and create a reference variable
        # extra row is so we can add the header row
        t = doc.add_table(df.shape[0] + 1, df.shape[1])

        # add the header rows.
        for j in range(df.shape[-1]):
            t.cell(0, j).text = df.columns[j]

        # add the rest of the data frame
        for i in range(df.shape[0]):
            for j in range(df.shape[-1]):
                t.cell(i + 1, j).text = str(df.values[i, j])

        # save the doc
        doc.save(doc_path)

    @staticmethod
    def save_df(path, file_name=None):
        import pandas as pd
        from tensorboard.backend.event_processing.event_accumulator import (
            EventAccumulator,
        )

        try:
            if file_name is not None:
                event_acc = EventAccumulator(join(path, file_name))
            else:
                event_acc = EventAccumulator(path)
            event_acc.Reload()
            tags = event_acc.Tags()["scalars"]
            values_df, columns = [], []
            for tag in tags:
                if tag == "epoch":
                    continue
                event_list = event_acc.Scalars(tag)
                values = list(map(lambda x: x.value, event_list))
                if "elapsed_time" not in columns:
                    wall_time = list(map(lambda x: x.wall_time, event_list))
                    wall_time = (
                        np.array(wall_time[1:]) - np.array(wall_time[:-1])
                    ).tolist()
                    wall_time = [np.mean(wall_time)] + wall_time
                    values_df.extend([values, wall_time])
                    columns.extend([f"{tag}", f"elapsed_time"])
                else:
                    values_df.append(values)
                    columns.append(tag)
        except Exception:
            print("Event file possibly corrupt: {}".format(path))
            traceback.print_exc()
        # appending epoch
        values_df.insert(0, list(range(1, len(values_df[0]) + 1)))
        columns.insert(0, "epoch")
        values = np.round(np.array(values_df).T, 3)
        df = pd.DataFrame(values, columns=columns)
        df = df[
            [
                "epoch",
                "elapsed_time",
                "train_f1_score",
                "train_loss",
                "train_acc",
                "val_f1_score",
                "val_loss",
                "val_acc",
            ]
        ]
        df_path = join(path, "csv_log.csv")
        df.to_csv(df_path, index=False)
        if "elapsed_time" in df:
            print("mean time:", df["elapsed_time"].mean(axis=0))
            print("sum time:", df["elapsed_time"].sum(axis=0))

        print(f"[INFO] Successfully saved results to {df_path}")
        return df


if __name__ == "__main__":
    path = (
        "/home/ai/projects/car_recognition/output/exp_1/train/lightning_logs/version_0"
    )
    TensorboardUtils.save_df(path)



import time
from functools import wraps
from typing import Callable


def get_method_time(method) -> Callable:
    """
    Gets the elapsed time of a method
    :param method: The method that will be measured
    :return:
    """

    @wraps(method)
    def wrapper(self, *args, **kwargs):
        tic = time.time()
        results = method(self, *args, **kwargs)
        toc = time.time()
        elapsed_time = round(toc - tic, 4)
        print(
            f"elapsed time for {self.__class__.__name__}.{method.__name__}: {elapsed_time}"
        )
        return results

    return wrapper


def get_func_time(func) -> Callable:
    """
    Gets the elapsed time of a class
    :param func: The func that will be measured
    :return:
    """

    @wraps(func)
    def wrapper(*args, **kwargs):
        tic = time.time()
        results = func(*args, **kwargs)
        toc = time.time()
        elapsed_time = round(toc - tic, 4)
        print(f"elapsed time for {func.__name__}: {elapsed_time} seconds")
        return results

    return wrapper



from typing import Iterable, Union

from deep_utils.utils.logging_utils import log_print, value_error_log


def utm2wgs84(
    x: Union[float, Iterable],
    y: Union[float, Iterable],
    zone_number: int,
    northern: bool,
    zone_letter=None,
    strict=True,
    return_list=True,
    logger=None,
    verbose=0,
):
    import utm

    get_zero = False
    if isinstance(x, Iterable) and isinstance(y, Iterable):
        point_list = [(x_, y_) for x_, y_ in zip(x, y)]
    elif isinstance(x, float) and isinstance(y, float):
        point_list = [(x, y)]
        get_zero = True
    else:
        value_error_log(logger, "Input x and y are not supported!")

    new_points = [
        utm.to_latlon(
            x,
            y,
            zone_number=zone_number,
            northern=northern,
            strict=strict,
            zone_letter=zone_letter,
        )
        for x, y in point_list
    ]
    log_print(
        logger,
        f"Successfully projected input points from utm to lat-long",
        verbose=verbose,
    )
    if return_list:
        x, y = list(zip(*new_points))
        if get_zero:
            x, y = x[0], y[0]
        return x, y

    if get_zero:
        new_points = new_points[0]
    return new_points


def project_points(
    x: Union[float, Iterable],
    y: Union[float, Iterable],
    input_type="epsg:3857",
    out_type="wgs84",
    return_list=True,
    logger=None,
    verbose=0,
):
    """
    Changing the projection of input points.
    Args:
        x:
        y:
        input_type:
        out_type:
        return_list:
        logger:
        verbose:

    Returns:

    """
    from collections.abc import Iterable

    import geopandas as gpd
    from shapely.geometry import Point

    get_zero = False
    if isinstance(x, Iterable) and isinstance(y, Iterable):
        point_list = [Point(x_, y_) for x_, y_ in zip(x, y)]
    elif isinstance(x, float) and isinstance(y, float):
        point_list = [Point(x, y)]
        get_zero = True
    else:
        value_error_log(logger, "Input x and y are not supported!")

    in_points = gpd.GeoSeries(point_list, crs=input_type)
    output_points = in_points.to_crs(out_type)
    log_print(
        logger,
        f"Successfully projected input points from {input_type} to {out_type}",
        verbose=verbose,
    )
    output_points = [(point.x, point.y) for point in output_points]
    if return_list:
        x, y = list(zip(*output_points))
        if get_zero:
            x, y = x[0], y[0]
        return x, y

    if get_zero:
        output_points = output_points[0]

    return output_points



def easy_argparse(*args):
    """
    This is a simple implementation of arg_parser. Each argument should be passed as dictionaries
    Args:
        *args:

    Returns:

    """
    import argparse

    parser = argparse.ArgumentParser()

    for arg in args:
        tp = type(arg)
        if tp == dict:
            name = arg.pop("name")
            parser.add_argument(name, **arg)
        else:
            raise TypeError(f"easy_argparse just supports dictionaries")
    return parser.parse_args()



import os
import sys
from functools import wraps

import requests

from deep_utils.utils.logging_utils import log_print, value_error_log


def get_file(fname, origin, cache_dir=None):
    base_dir, fname = os.path.split(fname)

    if cache_dir is None:
        if "DEEP_UTILS_HOME" in os.environ:
            cache_dir = os.environ.get("DEEP_UTILS_HOME")
        else:
            cache_dir = os.path.join(os.path.expanduser("~"), ".cache/deep_utils")

    datadir_base = os.path.expanduser(cache_dir)
    datadir = os.path.join(datadir_base, base_dir)
    os.makedirs(datadir, exist_ok=True)

    fpath = os.path.join(datadir, fname)
    if fpath.endswith(".zip"):
        check_path = fpath[:-4]
    else:
        check_path = fpath
    download = False if os.path.exists(check_path) else True

    if download:
        print("Downloading data from", origin)
        error_msg = "URL fetch failure on {}"
        try:
            with open(fpath, "wb") as f:
                response = requests.get(origin, stream=True)
                total = response.headers.get("content-length")

                if total is None:
                    f.write(response.content)
                else:
                    downloaded = 0
                    total = int(total)
                    for data in response.iter_content(
                            chunk_size=max(int(total / 1000), 1024 * 1024)
                    ):
                        downloaded += len(data)
                        f.write(data)
                        done = int(50 * downloaded / total)
                        sys.stdout.write(
                            "\r[{}{}]".format("█" * done, "." * (50 - done))
                        )
                        sys.stdout.flush()
            sys.stdout.write("\n")
        except (Exception, KeyboardInterrupt):
            if os.path.exists(fpath):
                os.remove(fpath)
            raise Exception(error_msg.format(origin))
        if fpath.endswith(".zip"):
            from zipfile import ZipFile

            try:
                with ZipFile(fpath, "r") as zip:
                    zip.printdir()
                    print(f"extracting {fpath}")
                    zip.extractall(fpath[:-4])
                    print(f"extracting is done!")
                os.remove(fpath)
            except (Exception, KeyboardInterrupt):
                raise Exception(f"Error in extracting {fpath}")
    if fpath.endswith(".zip"):
        fpath = fpath[:-4]
    return fpath


def download_file(
        url,
        download_dir=".",
        unzip=False,
        remove_zip=False,
        file_name=None,
        logger=None,
        remove_download=False,
):
    if url is None:
        value_error_log(logger, message="url is None. Exiting the function")

    os.makedirs(download_dir, exist_ok=True)
    log_print(logger, f"Downloading data from {url}")
    error_msg = "URL fetch failure on {}"
    download_des = None
    try:
        response = requests.get(url, stream=True)
        total = response.headers.get("content-length")
        while not total:
            response = requests.get(url, stream=True)
            total = response.headers.get("content-length")
        try:
            if file_name is None:
                file_name = response.headers.get("filename")
                if file_name is None:
                    file_name = response.headers.get("content-disposition").split("=")[
                        -1
                    ]
        except:
            file_name = os.path.split(url)[-1]

        download_des = os.path.join(download_dir, file_name)
        with open(download_des, "wb") as f:
            if total is None:
                f.write(response.content)
            else:
                downloaded = 0
                total = int(total)
                for data in response.iter_content(
                        chunk_size=max(int(total / 1000), 1024 * 1024)
                ):
                    downloaded += len(data)
                    f.write(data)
                    done = int(50 * downloaded / total)
                    sys.stdout.write("\r[{}{}]".format(
                        "█" * done, "." * (50 - done)))
                    sys.stdout.flush()
        sys.stdout.write("\n")
    except (Exception, KeyboardInterrupt):
        if (
                download_des is not None
                and os.path.isfile(download_des)
                and remove_download
        ):
            log_print(logger, f"ًRemoving {download_des} due to error")
            os.remove(download_des)
        log_print(logger, error_msg.format(url), log_type="error")
        raise Exception(error_msg.format(url))
    if unzip:
        unzip_func(download_des, download_dir, remove_zip, logger)
    return download_des


def unzip_func(zip_file, res_dir, remove_zip=False, logger=None):
    """
    A file to unzip zip files!
    Args:
        zip_file:
        res_dir:
        remove_zip:
        logger:

    Returns:

    """
    from zipfile import ZipFile

    with ZipFile(zip_file, "r") as zip:
        zip.extractall(res_dir)
        log_print(logger, f"Successfully extracted {zip_file} to {res_dir}!")
    if zip_file is not None and remove_zip:
        os.remove(zip_file)
        log_print(logger, f"Successfully removed {zip_file}!")


def download_decorator(func):
    @wraps(func)
    def wrapper(self, *args, **kwargs):
        config = self.config
        download_variables = self.download_variables
        for var in download_variables:
            if getattr(config, var) is None:
                url = getattr(config, var + "_url")
                f_name = getattr(config, var + "_cache")
                f_path = get_file(f_name, url)
                setattr(config, var, f_path)
        return func(self, *args, **kwargs)

    return wrapper


if __name__ == "__main__":
    url = "https://isic-challenge-data.s3.amazonaws.com/2016/ISBI2016_ISIC_Part1_Test_GroundTruth.zip"
    download_file(url, "/home/ai/pooya", unzip=True, remove_zip=False)
    download_file(
        url, "/home/ai/pooya", unzip=True, remove_zip=False, file_name="wow.zip"
    )
    download_file(
        "https://github.com/Practical-AI/deep_utils/archive/refs/tags/0.4.2.zip",
        "/home/ai/pooya",
        unzip=True,
        remove_zip=True,
    )
    download_file(
        "https://github.com/Practical-AI/deep_utils/archive/refs/tags/0.4.2.zip",
        "/home/ai/pooya",
        unzip=False,
        remove_zip=True,
    )



def grouped_mean_std_bar_plot(
    x_names,
    mean,
    std,
    xlabel="",
    ylabel="",
    title="",
    legends="",
    width=0.25,
    save_path=None,
    ylim=(0, 1),
):
    import matplotlib.pyplot as plt
    import numpy as np

    x = np.arange(len(x_names))
    w = -len(mean) // 2 * (width / 2) if len(mean) % 2 == 0 else - \
        len(mean) // 2 * width
    plt.figure(figsize=(25, 12))
    # plot data in grouped manner of bar type
    colors = ["green", "orange"]
    for m, s, color, legend in zip(mean, std, colors, legends):
        plt.bar(x - w, m, width, color=color, label=legend)
        plt.errorbar(x - w, m, yerr=s, fmt="o", color="r")
        w += width
    plt.xticks(x, x_names)
    plt.xlabel(xlabel)
    plt.ylabel(ylabel)
    plt.ylim(ylim)
    plt.title(title)
    plt.legend(loc="best", fancybox=True, framealpha=0.5)
    if save_path is not None:
        plt.savefig(save_path, bbox_inches="tight")
    plt.show()



from typing import List, Tuple, Union


def check_bio_labels(input_labels: Union[str, List[str], Tuple[str]], query_label: str):
    """
    Checks whether the input-labels and queried labels are the same.
    :param input_labels:
    :param query_label:
    :return:
    >>> input_labels = ["B-Loc", "I-Loc"]
    >>> query_label = "Loc"
    >>> check_bio_labels(input_labels, query_label)
    >>> True
    """
    if isinstance(input_labels, str):
        input_labels = input_labels.replace("B-", "").replace("I-", "")
        if input_labels == query_label:
            return True
    elif isinstance(input_labels, list) or isinstance(input_labels, tuple):
        input_labels = [lbl.replace("B-", "").replace("I-", "")
                        for lbl in input_labels]
        if len(set(input_labels)) == 1 and input_labels[0] == query_label:
            return True
    else:
        raise ValueError("Input_labels's type is not supported!")
    return False


def combine_tokens_with_hashtag_ner(input_words: List[str], input_labels: List[str], ignore_labels_equality=False):
    """
    This function connects tokenized words with # to each other if they have the same labels. In case you want to
    connect words with # to each other without considering labels' equality, set ignore_labels_equality to True.
    :param input_words:
    :param input_labels:
    :param ignore_labels_equality: if set to True, ignores the labels' equality condition in the process of connecting
    terms with #s
    :return:
    >>> words =  ['تلفن', '98', '##21', '##55', '##87', '##14', '##42']
    >>> labels = [ 'B-phone', 'I-phone', 'I-phone', 'I-phone', 'I-phone', 'I-phone', 'I-phone']
    >>> combine_tokens_with_hashtag_ner(words, labels)
    (['تلفن', '982155871442'], ['B-phone', 'I-phone'])
    >>> words =  ['تلفن', '98', '##21', '##55', '##87', '##14', '##42'] + [","] + ['تلفن', '98', '##21', '##55', '##87', '##14', '##42']
    >>> labels = [ 'B-phone', 'I-phone', 'I-phone', 'I-phone', 'I-phone', 'I-phone', 'I-phone'] + ["o"] + [ 'B-phone', 'I-phone', 'I-phone', 'I-phone', 'I-phone', 'I-phone', 'I-phone']
    >>> combine_tokens_with_hashtag_ner(words, labels)
    (['تلفن', '982155871442', ',', 'تلفن', '982155871442'], ['B-phone', 'I-phone', 'o', 'B-phone', 'I-phone'])
    >>> # ignore_labels_equality samples
    >>> words =  ['تلفن', '98', '##21', '##55', '##87', '##14', '##42'] + [","] + ['تلفن', '98', '##21', '##55', '##87', '##14', '##42']
    >>> labels = [ 'B-phone', 'I-phone', 'I-phone', 'I-phone', 'I-phone', 'I-phone', 'I-phone'] + ["o"] + [ 'B-phone', 'I-phone', 'I-phone', 'I-phone', 'I-phone', 'I-phone', 'I-phone']
    >>> combine_tokens_with_hashtag_ner(words, labels, ignore_labels_equality=True)
    (['تلفن', '982155871442', ',', 'تلفن', '982155871442'], ['B-phone', 'I-phone', 'o', 'B-phone', 'I-phone'])
    >>> words = ["شهید", "##قندی", "نیلوفر", "خیابان","شهید","دکتر","بهشتی","خیابان","سهروردی","شمالی","پلاک", "497"]
    >>> labels = ["B-section", "B-street", "B-street", "B-street", "I-street", "I-street", "I-street", "B-street", "I-street", "I-street", "B-house_number", "I-house_number" ]
    >>> combine_tokens_with_hashtag_ner(words, labels, ignore_labels_equality=True)
    (['شهیدقندی', 'نیلوفر', 'خیابان', 'شهید', 'دکتر', 'بهشتی', 'خیابان', 'سهروردی', 'شمالی', 'پلاک', '497'], ['B-section', 'B-street', 'B-street', 'I-street', 'I-street', 'I-street', 'B-street', 'I-street', 'I-street', 'B-house_number', 'I-house_number'])
    """
    assert len(input_words) == len(input_labels), "[Error] Number of input_words and input_labels is not equal"
    res_words, res_labels = [], []
    sub_word = ""
    sub_label = None
    save_label = None
    for word, label in zip(input_words, input_labels):
        query_label = label.replace("I-", "").replace("B-", "")
        if "#" in word and ((query_label == sub_label or ignore_labels_equality) or not sub_label):
            sub_word += word.replace("#", "")
            sub_label = query_label
        else:
            if sub_word and save_label:
                res_words.append(sub_word)
                res_labels.append(save_label)
            sub_word = word
            sub_label = query_label
            save_label = label
    if sub_word and sub_label:
        res_words.append(sub_word)
        res_labels.append(save_label)
    return res_words, res_labels



import inspect
import os
from abc import ABC
from typing import Any, Dict, Union

from deep_utils.utils.lib_utils.main_utils import import_module
from deep_utils.utils.os_utils import split_all


class MainClass(ABC):
    def __init__(self, name, file_path, **kwargs):
        self.config = None
        self.model = None
        self.name = name
        self.download_variables: Union[tuple, None] = kwargs.get(
            "download_variables", None
        )
        self.load_config(file_path, **kwargs)
        self.load_model()

    def __repr__(self):
        message = f"{self.name} config attributes:"
        for name, val in inspect.getmembers(self.config):

            if not name.startswith("_"):
                if not inspect.ismethod(val):
                    message += f"\n{name} = {val}"
        return message

    def load_model(self):
        raise NotImplementedError("load_model is not implemented")

    def load_config(self, file_path, **kwargs):
        separated_files = os.path.join(
            "deep_utils", os.path.split(
                file_path)[0].split("deep_utils")[-1][1:]
        )
        file_ = ".".join(split_all(separated_files) + ["config"])
        config = import_module(file_, "Config")
        self.config = config()
        self.update_config(**kwargs)

    def __dict_update(self, out_dict: Dict[str, Any], in_dict: Dict[str, Any]):
        for arg, val in in_dict.items():
            if arg in out_dict:
                source = out_dict[arg]
                if isinstance(val, dict) and isinstance(source, dict):
                    target = self.__dict_update(source, val)
                    out_dict[arg] = target
                else:
                    out_dict[arg] = val
            else:
                out_dict[arg] = val

        return out_dict

    def update_config(self, **kwargs):
        for arg, val in kwargs.items():
            if hasattr(self.config, arg):
                if val is None:
                    continue
                source = getattr(self.config, arg)
                if isinstance(val, dict) and isinstance(source, dict):
                    target = self.__dict_update(source, val)
                    setattr(self.config, arg, target)
                else:
                    setattr(self.config, arg, val)
            else:
                raise ValueError(
                    f"[ERROR] the config file does not contain argument: {arg}"
                )



class MainConfig:
    def vars(self) -> dict:
        out = dict()
        for key in dir(self):
            val = getattr(self, key)
            if (key.startswith("__") and key.endswith("__")) or type(
                val
            ).__name__ == "method":
                continue
            else:
                out[key] = val
        return out

    def __repr__(self):
        configs = self.vars()
        view = f"{self.__class__.__name__} -> " + ", ".join(
            f"{k}: {v}" for k, v in configs.items()
        )
        return view



from deep_utils.utils.lib_utils import import_module

SSDCV2CaffeFaceDetector = import_module(
    "deep_utils.vision.face_detection.ssd.cv2.caffe.ssd_cv2_caffe_face_detection",
    "SSDCV2CaffeFaceDetector",
)



def dictnamedtuple(typename, field_names, *, rename=False, defaults=None, module=None):
    import sys as _sys
    from keyword import iskeyword as _iskeyword

    from _collections import _tuplegetter

    # Validate the field names.  At the user's option, either generate an error
    # message or automatically replace the field name with a valid name.
    if isinstance(field_names, str):
        field_names = field_names.replace(",", " ").split()
    field_names = list(map(str, field_names))
    typename = _sys.intern(str(typename))

    if rename:
        seen = set()
        for index, name in enumerate(field_names):
            if (
                not name.isidentifier()
                or _iskeyword(name)
                or name.startswith("_")
                or name in seen
            ):
                field_names[index] = f"_{index}"
            seen.add(name)

    for name in [typename] + field_names:
        if not isinstance(name, str):
            raise TypeError("Type names and field names must be strings")
        if not name.isidentifier():
            raise ValueError(
                "Type names and field names must be valid " f"identifiers: {name!r}"
            )
        if _iskeyword(name):
            raise ValueError(
                "Type names and field names cannot be a " f"keyword: {name!r}"
            )

    seen = set()
    for name in field_names:
        if name.startswith("_") and not rename:
            raise ValueError(
                "Field names cannot start with an underscore: " f"{name!r}"
            )
        if name in seen:
            raise ValueError(f"Encountered duplicate field name: {name!r}")
        seen.add(name)

    field_defaults = {}
    if defaults is not None:
        defaults = tuple(defaults)
        if len(defaults) > len(field_names):
            raise TypeError("Got more default values than field names")
        field_defaults = dict(
            reversed(list(zip(reversed(field_names), reversed(defaults))))
        )

    # Variables used in the methods and docstrings
    field_names = tuple(map(_sys.intern, field_names))
    num_fields = len(field_names)
    arg_list = repr(field_names).replace("'", "")[1:-1]
    repr_fmt = "(" + ", ".join(f"{name}=%r" for name in field_names) + ")"
    tuple_new = tuple.__new__
    _dict, _tuple, _len, _map, _zip = dict, tuple, len, map, zip

    # Create all the named tuple methods to be added to the class namespace

    s = f"def __new__(_cls, {arg_list}): return _tuple_new(_cls, ({arg_list}))"
    namespace = {"_tuple_new": tuple_new, "__name__": f"namedtuple_{typename}"}
    # Note: exec() has the side-effect of interning the field names
    exec(s, namespace)
    __new__ = namespace["__new__"]
    __new__.__doc__ = f"Create new instance of {typename}({arg_list})"
    if defaults is not None:
        __new__.__defaults__ = defaults

    @classmethod
    def _make(cls, iterable):
        result = tuple_new(cls, iterable)
        if _len(result) != num_fields:
            raise TypeError(
                f"Expected {num_fields} arguments, got {len(result)}")
        return result

    _make.__func__.__doc__ = (
        f"Make a new {typename} object from a sequence " "or iterable"
    )

    def __repr__(self):
        "Return a nicely formatted representation string"
        return self.__class__.__name__ + repr_fmt % self

    def _asdict(self):
        "Return a new dict which maps field names to their values."
        return _dict(_zip(self._fields, self))

    def __getnewargs__(self):
        "Return self as a plain tuple.  Used by copy and pickle."
        return _tuple(self)

    def _replace(self, /, **kwds):
        result = self._make(_map(kwds.pop, field_names, self))
        if kwds:
            raise ValueError(f"Got unexpected field names: {list(kwds)!r}")
        return result

    _replace.__doc__ = (
        f"Return a new {typename} object replacing specified " "fields with new values"
    )
    # Modify function metadata to help with introspection and debugging
    for method in (
        __new__,
        _make.__func__,
        _replace,
        __repr__,
        _asdict,
        __getnewargs__,
    ):
        method.__qualname__ = f"{typename}.{method.__name__}"

    # Build-up the class namespace dictionary
    # and use type() to build the result class
    class_namespace = {
        "__doc__": f"{typename}({arg_list})",
        "__slots__": (),
        "_fields": field_names,
        "_field_defaults": field_defaults,
        # alternate spelling for backward compatibility
        "_fields_defaults": field_defaults,
        "__new__": __new__,
        "_make": _make,
        "_replace": _replace,
        "__repr__": __repr__,
        "_asdict": _asdict,
        "__getnewargs__": __getnewargs__,
    }
    for index, name in enumerate(field_names):
        doc = _sys.intern(f"Alias for field number {index}")
        class_namespace[name] = _tuplegetter(index, doc)

    result = type(typename, (tuple,), class_namespace)

    # For pickling to work, the __module__ variable needs to be set to the frame
    # where the named tuple is created.  Bypass this step in environments where
    # sys._getframe is not defined (Jython for example) or sys._getframe is not
    # defined for arguments greater than 0 (IronPython), or where the user has
    # specified a particular module.
    if module is None:
        try:
            module = _sys._getframe(1).f_globals.get("__name__", "__main__")
        except (AttributeError, ValueError):
            pass
    if module is not None:
        result.__module__ = module

    class DictNamedTuple(result):

        DictNamedTuple = True
        TypeName = typename

        def keys(self):
            keys = []
            for key in self._fields:
                keys.append(key)
            return keys

        def items(self):
            found = []
            for key in self._fields:
                value = getattr(self, key)
                found.append((key, value))
            return found

        def values(self):
            values = []
            for key in self._fields:
                values.append(getattr(self, key))
            return values

        def get(self, key):
            value = getattr(self, key)
            return value

        def __getitem__(self, key):
            value = getattr(self, key)
            return value

    return DictNamedTuple



import os


def tf_set_seed(seed):
    import random

    import numpy as np
    import tensorflow as tf

    os.environ["PYTHONHASHSEED"] = "0"

    tf.random.set_seed(seed)
    np.random.seed(seed)
    random.seed(seed)

    os.environ["TF_DETERMINISTIC_OPS"] = "1"
    os.environ["TF_CUDNN_DETERMINISTIC"] = "1"

    tf.config.threading.set_inter_op_parallelism_threads(1)
    tf.config.threading.set_intra_op_parallelism_threads(1)



from enum import Enum
from typing import Sequence, Union
import numpy as np
from deep_utils.utils.logging_utils.logging_utils import log_print, value_error_log


class Point:
    class PointSource(Enum):
        Torch = "Torch"
        TF = "TF"
        CV = "CV"
        Numpy = "Numpy"

    @staticmethod
    def point2point(
            point,
            in_source: Union[str, PointSource],
            to_source: Union[str, PointSource],
            in_relative=None,
            to_relative=None,
            shape=None,
            shape_source=None,
    ):
        """
        >>> #The ability to write in_source and to_source in any mode (capital or small letters, etc.)

        >>> Point.point2point(point = [1, 5], in_source=Point.PointSource.Numpy, to_source=Point.PointSource.Torch)
        (5, 1)
        >>> Point.point2point(point = [1, 5], in_source="nUmpy", to_source=Point.PointSource.Torch)
        (5, 1)

        >>>  #Convert point from in_relative to to_relative

        >>> Point.point2point(point = [0.1, 0.05], shape = [10,100], in_relative = False, to_relative = True, in_source = 'numpy', to_source = 'NUMPY', shape_source = Point.PointSource.Numpy)
        Traceback (most recent call last):
         Point._point2point(point = [0.1, 0.05], shape=[10,100],in_relative=False,to_relative=True, in_source='numpy', to_source='NUMPY', shape_source=Point.PointSource.Numpy)
        ValueError: the input is  relative while in_relative is set to False
        >>> Point.point2point(point = [0.1, 0.05], shape = [10,100], in_relative = True, to_relative = False, in_source = 'TF', to_source = 'tF', shape_source = Point.PointSource.Numpy)
        [1.0, 5.0]
        >>> Point.point2point(point = [1, 5], shape = [10,100], in_relative = True, to_relative=False, in_source = 'numPY', to_source = 'NUMPy', shape_source = Point.PointSource.Numpy)
        Traceback (most recent call last):
         Point._point2point(point = [1, 5], shape = [10,100], in_relative = True, to_relative = False, in_source = 'numPY', to_source = 'NUMPy', shape_source = Point.PointSource.Numpy)
        ValueError: the input is not relative while in_relative is set to True
        >>> Point.point2point(point = [1, 5], shape = [10,100], in_relative = False, to_relative = True, in_source = 'NUMPY', to_source = 'NUmpy', shape_source = Point.PointSource.Numpy)
        [0.1, 0.05]
        """
        if point is None or len(point) == 0:
            pass
        elif isinstance(point[0], (tuple, list, np.ndarray)):
            point = [
                Point._point2point(
                    p,
                    in_source=in_source,
                    to_source=to_source,
                    in_relative=in_relative,
                    to_relative=to_relative,
                    shape=shape,
                    shape_source=shape_source,
                )
                for p in point
            ]
        else:
            point = Point._point2point(
                point,
                in_source=in_source,
                to_source=to_source,
                in_relative=in_relative,
                to_relative=to_relative,
                shape=shape,
                shape_source=shape_source,
            )
        return point

    @staticmethod
    def _point2point(
            point,
            in_source: Union[str, PointSource],
            to_source: Union[str, PointSource],
            in_relative=None,
            to_relative=None,
            shape=None,
            shape_source=None,
    ):
        if isinstance(in_source, Point.PointSource):
            in_source = in_source.value
        if isinstance(to_source, Point.PointSource):
            to_source = to_source.value
        in_source = in_source.lower()
        to_source = to_source.lower()

        if (
                in_source in [Point.PointSource.Torch.value.lower(),
                              Point.PointSource.CV.value.lower()]
                and to_source in [Point.PointSource.TF.value.lower(), Point.PointSource.Numpy.value.lower()]
        ) or (
                in_source in [Point.PointSource.TF.value.lower(),
                              Point.PointSource.Numpy.value.lower()]
                and to_source in [Point.PointSource.Torch.value.lower(), Point.PointSource.CV.value.lower()]
        ):
            point = (point[1], point[0])
        elif (
                (in_source is None and to_source is None)
                or in_source == to_source
                or (
                        in_source in [Point.PointSource.Torch.value.lower(),
                                      Point.PointSource.CV.value.lower()]
                        and to_source
                        in [Point.PointSource.CV.value, Point.PointSource.Torch.value.lower()]
                )
                or (
                        in_source in [Point.PointSource.TF.value.lower(),
                                      Point.PointSource.Numpy.value.lower()]
                        and to_source
                        in [Point.PointSource.TF.value.lower(), Point.PointSource.Numpy.value.lower()]
                )
        ):
            pass
        else:
            raise Exception(
                f"Conversion from {in_source} to {to_source} is not Supported."
                f" Supported types: {Box._get_enum_names(Point.PointSource)}"
            )
        if to_source is not None and shape_source is not None and shape is not None:
            img_w, img_h = Point.point2point(
                shape, in_source=shape_source, to_source=to_source
            )
            if not in_relative:
                if isinstance(point[0], float) or isinstance(point[1], float):
                    raise ValueError(f"the input is  relative while in_relative is set to {in_relative}")
            if in_relative:
                if isinstance(point[0], int) or isinstance(point[1], int):
                    raise ValueError(f"the input is not relative while in_relative is set to {in_relative}")
            if not in_relative and to_relative:
                p1, p2 = point
                point = [p1 / img_w, p2 / img_h]
            elif in_relative and not to_relative:
                p1, p2 = point
                point = [p1 * img_w, p2 * img_h]
        return point

    @staticmethod
    def _put_point(
            img,
            point,
            radius,
            color=(0, 255, 0),
            thickness=None,
            lineType=None,
            shift=None,
            in_source="Numpy",
    ):
        import cv2

        if not isinstance(point, int):
            point = (int(point[0]), int(point[1]))
        point = Point.point2point(point, in_source=in_source, to_source="CV")
        return cv2.circle(img, point, radius, color, thickness, lineType, shift)

    @staticmethod
    def put_point(
            img,
            point,
            radius,
            color=(0, 255, 0),
            thickness=None,
            lineType=None,
            shift=None,
            in_source="Numpy",
    ):
        if point is None or len(point) == 0:
            pass
        elif isinstance(point[0], (tuple, list, np.ndarray)):
            for p in point:
                img = Point._put_point(
                    img, p, radius, color, thickness, lineType, shift, in_source
                )
        else:
            img = Point._put_point(
                img, point, radius, color, thickness, lineType, shift, in_source
            )
        return img

    @staticmethod
    def sort_points(pts: Union[list, tuple]):
        """
        Sort a list of 4 points based on upper-left, upper-right, down-right, down-left
        :param pts:
        :return:
        """
        top_points = sorted(pts, key=lambda l: l[0])[:2]
        top_left = min(top_points, key=lambda l: l[1])
        top_right = max(top_points, key=lambda l: l[1])
        pts.remove(top_left)
        pts.remove(top_right)
        down_left = min(pts, key=lambda l: l[1])
        down_right = max(pts, key=lambda l: l[1])
        return top_left, top_right, down_right, down_left


class Box:
    class BoxFormat(Enum):
        XYWH = "XYWH"
        XYXY = "XYXY"
        XCYC = "XCYC"

    class BoxSource(Enum):
        Torch = "Torch"
        TF = "TF"
        CV = "CV"
        Numpy = "Numpy"

    class OutType(Enum):
        Numpy = np.array
        List = list
        Tuple = tuple

    @staticmethod
    def box2box(
            box,
            in_format=None,
            to_format=None,
            in_source: Union[str, BoxSource] = BoxSource.Numpy,
            to_source: Union[str, BoxSource] = BoxSource.Numpy,
            in_relative=None,
            to_relative=None,
            shape=None,
            shape_source=None,
            out_type=None,
            return_int=None,
    ):
        if box is None or len(box) == 0:
            pass
        elif isinstance(box[0], (tuple, list, np.ndarray)):
            box = [
                Box._box2box(
                    b,
                    in_format=in_format,
                    to_format=to_format,
                    in_source=in_source,
                    to_source=to_source,
                    in_relative=in_relative,
                    to_relative=to_relative,
                    shape=shape,
                    shape_source=shape_source,
                    out_type=out_type,
                    return_int=return_int,
                )
                for b in box
            ]

        else:
            box = Box._box2box(
                box,
                in_format=in_format,
                to_format=to_format,
                in_source=in_source,
                to_source=to_source,
                in_relative=in_relative,
                to_relative=to_relative,
                shape=shape,
                shape_source=shape_source,
                out_type=out_type,
                return_int=return_int,
            )
        return box

    @staticmethod
    def _box2box(
            box,
            in_format=None,
            to_format=None,
            in_source=None,
            to_source=None,
            in_relative=None,
            to_relative=None,
            shape=None,
            shape_source: Point.PointSource = None,
            out_type=None,
            return_int=None,
            logger=None
    ):
        """

        :param box:
        :param in_format:
        :param to_format:
        :param in_source:
        :param to_source:
        :param in_relative:
        :param shape:
        :param to_relative:
        :param out_type: output type of the box. Supported types: list, tuple, numpy
        :param logger:
        :return:
        """
        if isinstance(in_format, Box.BoxFormat):
            in_format = in_format.value
        if isinstance(to_format, Box.BoxFormat):
            to_format = to_format.value

        if isinstance(in_source, Box.BoxSource):
            in_source = in_source.value
        if isinstance(to_source, Box.BoxSource):
            to_source = to_source.value

        if (
                in_format == Box.BoxFormat.XYWH.value
                and to_format == Box.BoxFormat.XYXY.value
        ):
            x1, y1, w, h = box
            x2, y2 = x1 + w, y1 + h
            box = [x1, y1, x2, y2]
        elif (
                in_format == Box.BoxFormat.XYXY.value
                and to_format == Box.BoxFormat.XYWH.value
        ):
            x1, y1, x2, y2 = box
            w, h = x2 - x1, y2 - y1
            box = [x1, y1, w, h]
        elif (
                in_format == Box.BoxFormat.XYXY.value
                and to_format == Box.BoxFormat.XCYC.value
        ):
            x1, y1, x2, y2 = box
            w, h = x2 - x1, y2 - y1
            xc, yc = (x1 + x2) / 2, (y1 + y2) / 2
            box = [xc, yc, w, h]
        elif (
                in_format == Box.BoxFormat.XCYC.value
                and to_format == Box.BoxFormat.XYXY.value
        ):
            xc, yc, w, h = box
            x1, y1, x2, y2 = xc - w / 2, yc - h / 2, xc + w / 2, yc + h / 2
            box = [x1, y1, x2, y2]
        elif (
                in_format == Box.BoxFormat.XYWH.value
                and to_format == Box.BoxFormat.XCYC.value
        ):
            x1, y1, w, h = box
            x2, y2 = x1 + w, y1 + h
            xc, yc = (x1 + x2) / 2, (y1 + y2) / 2
            box = [xc, yc, w, h]
        elif (
                in_format == Box.BoxFormat.XCYC.value
                and to_format == Box.BoxFormat.XYWH.value
        ):
            xc, yc, w, h = box
            x1, y1 = xc - w // 2, yc - h // 2

            box = [x1, y1, w, h]
        elif (in_format is None and to_format is None) or in_format == to_format:
            pass
        else:
            raise Exception(
                f"Conversion form {in_format} to {to_format} is not Supported."
                f" Supported types: {Box._get_enum_names(Box.BoxFormat)}"
            )

        if (
                in_source in [Box.BoxSource.Torch.value, Box.BoxSource.CV.value]
                and to_source in [Box.BoxSource.TF.value, Box.BoxSource.Numpy.value]
        ) or (
                in_source in [Box.BoxSource.TF.value, Box.BoxSource.Numpy.value]
                and to_source in [Box.BoxSource.Torch.value, Box.BoxSource.CV.value]
        ):
            box = [box[1], box[0], box[3], box[2]]
        elif (
                (in_source is None and to_source is None)
                or in_source == to_source
                or (
                        in_source in [Box.BoxSource.Torch.value,
                                      Box.BoxSource.CV.value]
                        and to_source in [Box.BoxSource.CV.value, Box.BoxSource.Torch.value]
                )
                or (
                        in_source in [Box.BoxSource.TF.value,
                                      Box.BoxSource.Numpy.value]
                        and to_source in [Box.BoxSource.TF.value, Box.BoxSource.Numpy.value]
                )
        ):
            pass
        else:
            raise Exception(
                f"Conversion form {in_source} to {to_source} is not Supported."
                f" Supported types: {Box._get_enum_names(Box.BoxSource)}"
            )
        if in_relative != to_relative:
            if to_source is not None and shape_source is not None and shape is not None:
                img_w, img_h = Point.point2point(
                    shape, in_source=shape_source, to_source=to_source
                )
                if not in_relative and to_relative:
                    b1, b2, b3, b4 = box
                    box = [b1 / img_w, b2 / img_h, b3 / img_w, b4 / img_h]
                elif in_relative and not to_relative:
                    b1, b2, b3, b4 = box
                    box = [b1 * img_w, b2 * img_h, b3 * img_w, b4 * img_h]
            else:
                value_error_log(logger,
                                "to_source, shape_source, and shape should contain values and should not be None")

        box = Box.get_type(box, out_type)
        if return_int:
            box = [int(b) for b in box]
        return box

    @staticmethod
    def get_type(in_, out_type):
        if out_type is not None:
            try:
                in_ = out_type(in_)
            except:
                raise Exception(
                    f"{out_type} is not Supported. Supported types: {Box._get_enum_names(Box.OutType)}"
                )
        return in_

    @staticmethod
    def get_enum_names(in_):
        return [n.name for n in in_]

    @staticmethod
    def _put_box_pil(img,
                     box,
                     outline="green",
                     fill=None,
                     in_relative=False,
                     in_format="XYXY",
                     in_source="Numpy",
                     return_numpy=True):
        from PIL import ImageDraw, Image
        if isinstance(img, np.ndarray):
            img = Image.fromarray(img)
        box = Box.box2box(
            box,
            in_format=in_format,
            to_format=Box.BoxFormat.XYXY,
            in_source=in_source,
            to_source=Box.BoxSource.Numpy,
            in_relative=in_relative,
            to_relative=False,
            shape=img.size,
            shape_source="Numpy",
        )

        shape = ((box[0], box[1]), (box[2], box[3]))
        draw = ImageDraw.Draw(img)
        draw.rectangle(shape, fill=fill, outline=outline)
        if return_numpy:
            np.array(img)
        return img

    @staticmethod
    def put_box_pil(
            img,
            box,
            outline="green",
            fill=None,
            in_relative=False,
            in_format="XYXY",
            in_source="Numpy",
            return_numpy=True
    ):
        if box is None or len(box) == 0:
            pass
        elif isinstance(box[0], (tuple, list, np.ndarray)):
            for b in box:
                img = Box._put_box_pil(
                    img=img,
                    box=b,
                    outline=outline,
                    fill=fill,
                    in_relative=in_relative,
                    in_format=in_format,
                    in_source=in_source,
                    return_numpy=return_numpy)
        else:
            img = Box._put_box_pil(
                img=img,
                box=box,
                outline=outline,
                fill=fill,
                in_relative=in_relative,
                in_format=in_format,
                in_source=in_source,
                return_numpy=return_numpy)

        return img

    @staticmethod
    def _put_box(
            img,
            box,
            copy=False,
            color=(0, 255, 0),
            thickness=1,
            lineType=None,
            shift=None,
            in_relative=False,
            in_format="XYXY",
            in_source="Numpy",
    ):

        import cv2

        box = Box.box2box(
            box,
            in_format=in_format,
            to_format=Box.BoxFormat.XYXY,
            in_source=in_source,
            to_source=Box.BoxSource.CV,
            in_relative=in_relative,
            to_relative=False,
            shape=img.shape[:2],
            shape_source="Numpy",
        )

        if not isinstance(img, np.ndarray):
            img = np.array(img).astype(np.uint8)
        else:
            img = img.astype(np.uint8)
        box = [int(point) for point in box]
        if copy:
            img = img.copy()
        img = cv2.rectangle(
            img,
            (box[0], box[1]),
            (box[2], box[3]),
            color=color,
            thickness=thickness,
            lineType=lineType,
            shift=shift,
        )
        return img

    @staticmethod
    def put_box(
            img,
            box,
            copy=False,
            color=(0, 255, 0),
            thickness=1,
            lineType=None,
            shift=None,
            in_relative=False,
            in_format=BoxFormat.XYXY,
            in_source=BoxSource.Numpy,
    ):
        if box is None or len(box) == 0:
            pass
        elif isinstance(box[0], (tuple, list, np.ndarray)):
            for b in box:
                img = Box._put_box(
                    img,
                    box=b,
                    copy=copy,
                    color=color,
                    thickness=thickness,
                    lineType=lineType,
                    shift=shift,
                    in_format=in_format,
                    in_source=in_source,
                    in_relative=in_relative,
                )
        else:
            img = Box._put_box(
                img,
                box=box,
                copy=copy,
                color=color,
                thickness=thickness,
                lineType=lineType,
                shift=shift,
                in_format=in_format,
                in_source=in_source,
                in_relative=in_relative,
            )

        return img

    @staticmethod
    def _get_box_img(img, bbox, box_format=BoxFormat.XYXY, box_source=BoxSource.Numpy):
        bbox = Box.box2box(
            bbox,
            in_format=box_format,
            to_format=Box.BoxFormat.XYXY,
            in_source=box_source,
            to_source=Box.BoxSource.Numpy,
            return_int=True,
        )
        img_part = img[bbox[0]: bbox[2], bbox[1]: bbox[3]]
        return img_part

    @staticmethod
    def get_box_img(img, bbox, box_format=BoxFormat.XYXY, box_source=BoxSource.Numpy):
        if len(img.shape) != 3:
            raise Exception("The image size should be 3")

        img_part = []
        if bbox is None or len(bbox) == 0:
            pass
        elif isinstance(bbox[0], (tuple, list, np.ndarray)):
            img_part = [Box._get_box_img(
                img, b, box_format, box_source) for b in bbox]
        else:
            img_part = Box._get_box_img(img, bbox, box_format, box_source)
        return img_part

    @staticmethod
    def _put_text_pil(img, text, org, color=(0, 255, 0), font=None, font_size=32, return_np=True):
        from PIL import Image, ImageFont, ImageDraw

        if isinstance(img, np.ndarray):
            img = Image.fromarray(img)
        if isinstance(font, str):
            font = ImageFont.truetype(font, font_size)

        draw = ImageDraw.Draw(img)
        draw.text(org, text, color, font=font)
        if return_np:
            img = np.array(img)
        return img

    @staticmethod
    def put_text_pil(
            img,
            text,
            org,
            color=(0, 255, 0),
            font=None,
            font_size: int = 32,
            return_np=True
    ):
        if text is None or len(text) == 0 or org is None or len(org) == 0:
            pass
        elif isinstance(text, (tuple, list, np.ndarray)):
            for t, o in zip(text, org):
                img = Box._put_text_pil(img, t, o, color, font, font_size, return_np)
        else:
            img = Box._put_text_pil(img, text, org, color, font, font_size, return_np)
        return img

    @staticmethod
    def _put_text(
            img,
            text,
            org,
            fontFace=None,
            fontScale=1,
            color=(0, 255, 0),
            thickness=1,
            lineType=None,
            bottomLeftOrigin=None,
            org_source="Numpy"
    ):
        import cv2

        org = (int(org[0]), int(org[1]))
        org = Point.point2point(
            org, in_source=org_source, to_source=Point.PointSource.CV
        )
        font_face = cv2.FONT_HERSHEY_PLAIN if fontFace is None else fontFace
        img = cv2.putText(
            img,
            text,
            org,
            font_face,
            fontScale,
            color,
            thickness,
            lineType,
            bottomLeftOrigin,
        )

        return img

    @staticmethod
    def put_text(
            img,
            text,
            org,
            fontFace=None,
            fontScale: float = 1,
            color=(0, 255, 0),
            thickness=1,
            lineType=None,
            bottomLeftOrigin=None,
            org_source="Numpy",
    ):
        if text is None or len(text) == 0 or org is None or len(org) == 0:
            pass
        elif isinstance(text, (tuple, list, np.ndarray)):
            for t, o in zip(text, org):
                img = Box._put_text(
                    img,
                    t,
                    o,
                    fontFace,
                    fontScale,
                    color,
                    thickness,
                    lineType,
                    bottomLeftOrigin,
                    org_source=org_source,
                )
        else:
            img = Box._put_text(
                img,
                text,
                org,
                fontFace,
                fontScale,
                color,
                thickness,
                lineType,
                bottomLeftOrigin,
                org_source=org_source,
            )
        return img

    @staticmethod
    def get_biggest(
            box,
            in_format=BoxFormat.XYXY,
            in_source=BoxSource.Numpy,
            get_index=False,
            inputs: Union[None, dict] = None,
            reverse=False,
    ):
        if len(box) == 0 or box is None:
            return
        box = Box.box2box(
            box,
            in_format=in_format,
            in_source=in_source,
            to_source=Box.BoxSource.Numpy,
            to_format=Box.BoxFormat.XYWH,
        )
        if reverse:
            chosen_box = min(box, key=lambda b: b[2] * b[3])
        else:
            chosen_box = max(box, key=lambda b: b[2] * b[3])
        index = box.index(chosen_box)
        if inputs is not None:
            inputs = {k: v[index] for k, v in inputs.items()}
            return inputs
        chosen_box = Box.box2box(
            chosen_box, in_format=Box.BoxFormat.XYWH, to_format=Box.BoxFormat.XYXY
        )
        if get_index:
            return chosen_box, index
        return chosen_box

    @staticmethod
    def get_area(box, in_format=BoxFormat.XYXY, in_source=BoxSource.Numpy):
        box = Box.box2box(
            box,
            in_format=in_format,
            in_source=in_source,
            to_source=Box.BoxSource.Numpy,
            to_format=Box.BoxFormat.XYWH,
        )
        area = box[2] * box[3]
        return area

    @staticmethod
    def fill_box(img, box, value, in_format=BoxFormat.XYXY, in_source=BoxSource.Numpy):
        """
        Fill the selected box with the specified value.
        :param img: The input image
        :param box: the box that should be filled
        :param value: the value with which the box will be filled
        :param in_format: box input format
        :param in_source: box input source
        :return: the filled box
        """
        bbox = Box.box2box(
            box,
            in_format=in_format,
            in_source=in_source,
            to_source=Box.BoxSource.Numpy,
            to_format=Box.BoxFormat.XYXY,
        )
        img[bbox[0]: bbox[2], bbox[1]: bbox[3]] = value
        return img

    @staticmethod
    def fill_outer_box(
            img, box, value: int = 0, in_format=BoxFormat.XYXY, in_source=BoxSource.Numpy
    ):
        """
        Fill the outer area of the selected box with the specified value.
        :param img: The input image
        :param box: the box that should remain fixed
        :param value: the value with which the outer box will be filled, default is zero
        :param in_format: box input format
        :param in_source: box input source
        :return: the filled box
        """
        import cv2

        bbox = Box.box2box(
            box,
            in_format=in_format,
            in_source=in_source,
            to_source=Box.BoxSource.Numpy,
            to_format=Box.BoxFormat.XYXY,
        )
        mask = np.ones_like(img, dtype=np.uint8) * value
        mask[bbox[0]: bbox[2], bbox[1]: bbox[3]] = 1
        img = cv2.multiply(img, mask)
        return img

    @staticmethod
    def _put_box_text_pil(img, box, label, outline, text_color, text_font_size, text_font, return_numpy=True):
        from PIL import ImageDraw, Image, ImageFont
        if isinstance(img, np.ndarray):
            pil_img = Image.fromarray(img)
        else:
            pil_img = img
        if text_font is None:
            pil_font = ImageFont.load_default()
        elif isinstance(text_font, str):
            pil_font = ImageFont.truetype(text_font, text_font_size)
        p1, p2 = (int(box[0]), int(box[1])), (int(box[2]), int(box[3]))
        pil_img = Box.put_box_pil(pil_img, [*p1, *p2], outline=outline, return_numpy=return_numpy)

        draw_interface = ImageDraw.Draw(pil_img)
        w, h = draw_interface.textsize(label, pil_font)
        outside = p1[1] - h - 3 >= 0  # label fits outside box
        p2 = p1[0] + w, (p1[1] - h - 3) if outside else (p1[1] + h + 3)
        pil_img = Box.put_box_pil(pil_img, [*p1, *p2], outline=outline, fill=outline, return_numpy=return_numpy)

        org = p1[0], (p1[1] - h - 2) if outside else (p1[1] + h + 2)

        pil_img = Box.put_text_pil(pil_img, label, org, color=text_color, font_size=text_font_size, font=pil_font,
                                   return_np=return_numpy)

        img = np.array(pil_img) if return_numpy else pil_img
        return img

    @staticmethod
    def put_box_text_pil(
            img: Union[Sequence, np.ndarray],
            box: Union[Sequence],
            label: Union[Sequence, str],
            outline=(128, 128, 128),
            txt_color=(255, 255, 255),
            text_font_size=32,
            text_font=None,
            return_numpy=True,
            in_source=BoxSource.CV,
    ):
        box = Box.box2box(
            box, in_source=in_source, to_source=Box.BoxSource.Numpy
        )
        if (
                isinstance(box, Sequence)
                and isinstance(box[0], Sequence)
                and isinstance(label, Sequence)
        ):
            if isinstance(outline, Sequence) and isinstance(outline[0], Sequence):
                if isinstance(txt_color, Sequence) and isinstance(txt_color[0], Sequence):
                    for b, l, c, t_c in zip(box, label, outline, txt_color):
                        img = Box._put_box_text_pil(img, b, l, c, t_c, text_font_size, text_font, return_numpy)
                else:
                    for b, l, c in zip(box, label, outline):
                        img = Box._put_box_text_pil(img, b, l, c, txt_color, text_font_size, text_font,
                                                    return_numpy)
            else:
                for b, l in zip(box, label):
                    img = Box._put_box_text_pil(img, b, l, outline, txt_color, text_font_size, text_font,
                                                return_numpy)
        else:
            img = Box._put_box_text_pil(img, box, label, outline, txt_color, text_font_size, text_font,
                                        return_numpy)
        return img

    @staticmethod
    def _put_box_text(
            img, box, label, color=(128, 128, 128), txt_color=(255, 255, 255), thickness=2
    ):
        import cv2

        p1, p2 = (int(box[0]), int(box[1])), (int(box[2]), int(box[3]))
        img = Box.put_box(
            img,
            box,
            color=color,
            thickness=thickness,
            lineType=cv2.LINE_AA,
            in_source=Box.BoxSource.CV,
        )
        text_font = max(thickness - 1, 1)  # font thickness
        w, h = cv2.getTextSize(label, 0, fontScale=thickness / 3, thickness=text_font)[
            0
        ]  # text width, height
        outside = p1[1] - h - 3 >= 0  # label fits outside box
        p2 = p1[0] + w, p1[1] - h - 3 if outside else p1[1] + h + 3
        img = Box.put_box(
            img,
            [*p1, *p2],
            color=color,
            thickness=-1,
            lineType=cv2.LINE_AA,
            in_source=Box.BoxSource.CV,
        )
        x0, x1 = p1[0], p1[1] - 2 if outside else p1[1] + h + 2
        img = Box.put_text(
            img,
            label,
            (x0, x1),
            fontFace=0,
            fontScale=thickness / 3,
            color=txt_color,
            thickness=text_font,
            lineType=cv2.LINE_AA,
            org_source="CV",
        )
        return img

    @staticmethod
    def put_box_text(
            img: Union[Sequence, np.ndarray],
            box: Union[Sequence],
            label: Union[Sequence, str],
            color=(128, 128, 128),
            txt_color=(255, 255, 255),
            thickness=2,
    ):
        """
        :param img:
        :param box: It should be in numpy source!
        :param label:
        :param color:
        :param txt_color:
        :param thickness:
        :return:
        """
        box = Box.box2box(
            box, in_source=Box.BoxSource.Numpy, to_source=Box.BoxSource.CV
        )
        if (
                isinstance(box, Sequence)
                and isinstance(box[0], Sequence)
                and isinstance(label, Sequence)
        ):
            if isinstance(color, Sequence) and isinstance(color[0], Sequence):
                if isinstance(txt_color, Sequence) and isinstance(
                        txt_color[0], Sequence
                ):
                    for b, l, c, t_c in zip(box, label, color, txt_color):
                        img = Box._put_box_text(img, b, l, c, t_c, thickness)
                else:
                    for b, l, c in zip(box, label, color):
                        img = Box._put_box_text(
                            img, b, l, c, txt_color, thickness)
            else:
                for b, l in zip(box, label):
                    img = Box._put_box_text(
                        img, b, l, color, txt_color, thickness)
        else:
            img = Box._put_box_text(
                img, box, label, color, txt_color, thickness)
        return img

    @classmethod
    def _get_enum_names(cls, PointSource):
        pass



try:
    from deep_utils.dummy_objects.vision.torch_vision import TorchVisionModel
    from deep_utils.vision.torch_vision.torch_vision_models.torch_vision_models import (
        TorchVisionModel,
    )
except:
    pass



import numpy as np


def shuffle_group(array_a: np.ndarray, array_b: np.ndarray = None, copy=False):
    """
    :param array_a: The input array. It could be a batch of images or ...
    :param array_b: The labels or the masks of the input array that should be shuffled together.
    :param copy: whether generate a copy of the inputs and apply the function or not
    :return:
    returns the shuffle format of array a and b if the latter one exists.
    """
    if copy:
        array_a = array_a.copy()
    if array_b is not None:
        if copy:
            array_b = array_b.copy()
        indices = np.arange(len(array_a))
        np.random.shuffle(indices)
        array_a[:] = array_a[indices]
        array_b[:] = array_b[indices]
        return array_a, array_b
    else:
        np.random.shuffle(array_a)
        return array_a


def shuffle_group_torch(array_a, array_b=None, copy=False):
    """
    :param array_a: The input array. It could be a batch of images or ...
    :param array_b: The labels or the masks of the input array that should be shuffled together.
    :param copy: whether generate a copy of the inputs and apply the function or not
    :return:
    returns the shuffle format of array a and b if the latter one exists.
    """
    import torch

    if copy:
        array_a = torch.clone(array_a)
    if array_b is not None:
        if copy:
            array_b = torch.clone(array_b)
        indices = np.arange(len(array_a))
        np.random.shuffle(indices)
        array_a[np.arange(len(array_a))] = array_a[indices]
        array_b[np.arange(len(array_a))] = array_b[indices]
        return array_a, array_b
    else:
        indices = np.arange(len(array_a))
        np.random.shuffle(indices)
        array_a[np.arange(len(array_a))] = array_a[indices]
        return array_a



import os
import shutil
import sys
import time
from os.path import join, isfile
from pathlib import Path
from typing import Dict, List, Type, Union

import numpy as np
import torch
from tqdm import tqdm

from deep_utils.vision.object_detection.yolo.yolo_detector import YOLOObjectDetector
from deep_utils.utils.box_utils.boxes import Box
from deep_utils.utils.dir_utils.dir_utils import (
    dir_train_test_split,
    remove_create,
    transfer_directory_items,
    file_incremental,
)
from deep_utils.utils.lib_utils.lib_decorators import (
    get_from_config,
    in_shape_fix,
    lib_rgb2bgr,
    out_shape_fix,
)
from deep_utils.utils.logging_utils import log_print
from deep_utils.utils.opencv_utils.main import show_destroy_cv2
from deep_utils.utils.os_utils.os_path import split_extension
from deep_utils.utils.dict_named_tuple_utils import dictnamedtuple
from deep_utils.utils.shutil_utils.shutil_utils import mv_or_copy

from .config import Config


# FILE = Path(__file__).resolve()
# ROOT = FILE.parents[0]  # YOLOv5 root directory
# if str(ROOT) not in sys.path:
#     sys.path.append(str(ROOT))  # add ROOT to PATH
# from .models.experimental import attempt_load
# from .utils_.datasets import letterbox
# from .utils_.general import non_max_suppression, scale_coords
#
# OUTPUT_CLASS = dictnamedtuple(
#     "Object", ["class_indices", "boxes", "confidences", "class_names", "elapsed_time"]
# )


class OutputType:
    class_indices: List[int]
    boxes: List[List[int]]
    confidences: List[float]
    class_names: List[str]
    elapsed_time: float


class YOLOV7TorchObjectDetector(YOLOObjectDetector):
    def __init__(
            self,
            class_names=None,
            model_weight=None,
            device="cpu",
            img_size=(320, 320),
            confidence=0.4,
            iou_thresh=0.45,
            **kwargs,
    ):
        super(YOLOV7TorchObjectDetector, self).__init__(
            name=self.__class__.__name__,
            file_path=__file__,
            class_names=class_names,
            model_weight=model_weight,
            device=device,
            img_size=img_size,
            confidence=confidence,
            iou_thresh=iou_thresh,
            **kwargs,
        )
        self.config: Config

    def load_model(self):
        pass



from pyannote.audio import Pipeline
from deep_utils.utils.logging_utils.logging_utils import log_print
from deep_utils.audio.audio_utils.torchaudio_utils import TorchAudioUtils
from deep_utils.utils.dir_utils.dir_utils import split_extension
import torchaudio
from pyannote.audio.pipelines import VoiceActivityDetection


class PyannoteAudioVAD:
    def __init__(self, device="cpu", min_duration_on=1, max_duration_on=10, logger=None):
        self.max_duration_on = max_duration_on
        self.vad_pipeline = VoiceActivityDetection(segmentation="pyannote/segmentation", device=device)
        self.vad_pipeline.instantiate({
            # onset/offset activation thresholds
            "onset": 0.5, "offset": 0.5,
            # remove speech regions shorter than that many seconds.
            "min_duration_on": min_duration_on,
            # fill non-speech regions shorter than that many seconds.
            "min_duration_off": 0
        })
        log_print(logger, "Successfully Created VAD!")
        self.speaker_segmentor_pipeline = Pipeline.from_pretrained("pyannote/speaker-segmentation")

    def __call__(self, wave, sr, logger=None, verbose=1):
        model_input = dict({"waveform": wave, "sample_rate": sr})
        speech_activities = self.vad_pipeline(model_input)
        audio_segments = []
        for i, speech_region in enumerate(speech_activities.get_timeline()):
            w = wave[0][int(speech_region.start * sr):int(speech_region.end * sr)].unsqueeze(0)
            ws = TorchAudioUtils.split(w, sr=sr, max_seconds=self.max_duration_on, logger=logger, verbose=verbose)
            audio_segments.extend(ws)
        log_print(logger, f"Successfully segmented wave-array to {len(audio_segments)} segments", verbose=verbose)
        return audio_segments

    def vad_save(self, wave_path, audio_dir=None, logger=None, verbose=1):
        wave, sr = torchaudio.load(wave_path)
        audio_segments = self(wave=wave, sr=sr, logger=logger, verbose=verbose)
        audio_dir = wave_path if audio_dir is None else audio_dir
        for index, segment in enumerate(audio_segments):
            audio_path = split_extension(audio_dir, suffix=f"_{index:03}")
            torchaudio.save(audio_path, segment, 16000)
        log_print(logger, f"Successfully saved wave to {audio_dir}", verbose=verbose)

# if __name__ == '__main__':
#     audio_vad = PyannoteAudioVAD()
#     audio_dirs = ["/home/ai/projects/speech/dataset/irancel-voice-dataset/old-audios/activation_segments",
#                   "/home/ai/projects/speech/dataset/irancel-voice-dataset/old-audios/general_segments",
#                   "/home/ai/projects/speech/dataset/irancel-voice-dataset/old-audios/repairs_segments",
#                   "/home/ai/projects/speech/dataset/irancel-voice-dataset/old-audios/arrival_segments"]
#     for audio_dir in audio_dirs:
#         for speech_name in os.listdir(audio_dir):
#             speech_path = os.path.join(audio_dir, speech_name)
#             for wave_name in os.listdir(speech_path):
#                 wave_path = os.path.join(speech_path, wave_name)
#                 audio_vad.vad_save(wave_path)



import sys

if sys.version_info.major >= 3 and sys.version_info.minor >= 8:
    from deep_utils.utils.dict_named_tuple_utils.dictnamedtuple_38 import dictnamedtuple
else:
    from deep_utils.utils.dict_named_tuple_utils.dictnamedtuple_37 import dictnamedtuple



import os
from pathlib import Path
from time import time
from typing import Union

import cv2
import numpy as np
import torch

from deep_utils.utils.logging_utils.logging_utils import log_print
from deep_utils.vision.torch_vision.torch_vision_models import TorchVisionModel
from deep_utils.vision.vision_utils.torch_vision_utils.torch_vision_utils import TorchVisionUtils


class TorchVisionInference:
    def __init__(self, model_path, device="cpu", logger=None, verbose=1):
        """
        A class for model inference
        :param model_path: Path to the model weights and params
        :param device: TorchVisionInference device
        :param logger:
        :param verbose:
        """
        self.logger = logger
        self.verbose = verbose
        save_params = torch.load(model_path, map_location=device)
        self.model = TorchVisionModel(
            model_name=save_params["model_name"],
            num_classes=save_params["n_classes"],
            last_layer_nodes=save_params["last_layer_nodes"],
            use_pretrained=False,
            feature_extract=True,
        )
        try:
            self.model.load_state_dict(save_params["state_dict"])
        except:
            self.model.load_state_dict(
                {
                    ".".join(k.split(".")[1:]): v
                    for k, v in save_params["state_dict"].items()
                }
            )

        self.device = device
        self.label_map = save_params["id_to_class"]
        self.model.eval()
        self.transform = save_params["val_transform"]

    def infer(self, img: Union[str, Path, np.ndarray], return_confidence=False):
        """

        :param img:
        :param return_confidence: If True, return confidence alongside the prediction class
        :return:
        """
        if isinstance(img, Path) or isinstance(img, str):
            img = cv2.imread(img)[..., ::-1]
        image = self.transform(image=img)["image"]
        image = image.view(1, *image.size())
        with torch.no_grad():
            image = image.to(self.device)
            logits = self.model(image).squeeze(0)
        confidence, prediction = torch.max(torch.softmax(logits, dim=0), dim=0)
        confidence, prediction = confidence.item(), prediction.item()
        label = self.label_map[prediction]
        if return_confidence:
            return label, confidence
        return self.label_map[prediction]

    def infer_group(self, images: np.ndarray, return_confidence=False):
        """
        Infer a batch of images
        :param images: A batch of images
        :param return_confidence: If True, return confidence alongside the prediction class
        :return:
        """
        images_tensor = TorchVisionUtils.transform_concatenate_images(images, self.transform, device=self.device)
        with torch.no_grad():
            logits = self.model(images_tensor)
        confidence, prediction = torch.max(torch.softmax(logits, dim=1), dim=1)
        labels = [self.label_map[lbl.item()] for lbl in prediction]
        if return_confidence:
            return labels, confidence.numpy()
        else:
            return labels

    def infer_directory(self, dir_path: Union[Path, str]):
        """
        Infer a directory of images!
        :param dir_path:
        :return:
        """
        for image_name in sorted(os.listdir(dir_path)):
            image_path = os.path.join(dir_path, image_name)
            try:
                tic = time()
                prediction = self.infer(image_path)
                toc = time()
                log_print(
                    self.logger,
                    f"predicted class for {image_path} is {prediction}\ninference time: {toc - tic}",
                    verbose=self.verbose,
                )
            except BaseException as e:
                log_print(
                    self.logger,
                    f"img: {image_path} is invalid -> {e}",
                    verbose=self.verbose,
                )



from deep_utils.utils.lib_utils.main_utils import list_utils, loader

from .main import ObjectDetector
from .yolo import *

Object_Detection_Models = {"YOLOV5TorchObjectDetector": YOLOV5TorchObjectDetector,
                           "YOLOV7TorchObjectDetector": YOLOV7TorchObjectDetector}

list_object_detection_models = list_utils(Object_Detection_Models)


def object_detector_loader(name, **kwargs) -> ObjectDetector:
    return loader(Object_Detection_Models, list_object_detection_models)(name, **kwargs)



import logging
from typing import Union

import numpy as np

from deep_utils.utils.logging_utils import log_print, value_error_log


def img_to_b64(image: np.ndarray, extension: str = ".jpg") -> str:
    """
    returns a base64 encoded string from an image
    :param image: The input image
    :param extension: the extension to encode
    :return: byte string
    """
    import base64

    import cv2

    _, encoded_img = cv2.imencode(extension, image)
    base64_img = base64.b64encode(encoded_img).decode("utf-8")
    return base64_img


def b64_to_img(image_byte: Union[bytes, str]) -> np.ndarray:
    """
    Converts the input bytes or string to an 3-channel image
    :param image_byte: base64 image string
    :return: numpy image
    """
    import base64
    import cv2
    if isinstance(image_byte, bytes):
        image_byte = image_byte.decode()
    if ";" in image_byte:
        image_byte = image_byte.split(";")[-1]

    image_byte = image_byte.encode()
    im_bytes = base64.b64decode(image_byte)
    im_arr = np.frombuffer(im_bytes, dtype=np.uint8)  # im_arr is one-dim Numpy array
    img = cv2.imdecode(im_arr, flags=cv2.IMREAD_COLOR)

    return img


def ndarray_to_b64(
        array: np.ndarray,
        dtype: Union[None, str, type] = None,
        append_shape=False,
        append_dtype=False,
        decode=None,
        logger: Union[None, logging.Logger] = None,
):
    """
    Converting a ndarray to base64. For images use "img_to_b64"
    :param array:
    :param dtype:
    :return:
    """
    import base64

    if dtype is None:
        dtype = array.dtype
    array = array.astype(dtype)
    shape = array.shape
    shape_bytes = base64.struct.pack(
        f">6s{len(shape) + 1}I", bytes("shape:",
                                       "utf-8"), len(shape) * 4, *shape
    )
    array_bytes = base64.b64encode(array)
    if isinstance(dtype, str):
        dtype_name = dtype
    elif isinstance(dtype, type):
        dtype_name = dtype.__name__
    elif isinstance(dtype, np.dtype):
        dtype_name = dtype.name
    else:
        value_error_log(logger, f"dtype: {dtype} is not supported!")
    dtype_bytes = base64.struct.pack(
        f">6sI{len(dtype_name)}s",
        bytes("dtype:", "utf-8"),
        len(dtype_name),
        bytes(dtype_name, "utf-8"),
    )
    if not append_dtype and not append_dtype:
        res = array_bytes
    elif not append_dtype and append_shape:
        res = shape_bytes + array_bytes
    elif append_dtype and not append_shape:
        res = dtype_bytes + array_bytes
    elif append_dtype and append_shape:
        res = dtype_bytes + shape_bytes + array_bytes
    else:
        value_error_log(
            logger, f"dtype: {dtype}, shape: {shape} is not supported")

    if decode:
        res = res.decode(decode)

    return res


def b64_to_ndarray(
        byte_array, dtype, shape, logger: Union[None, logging.Logger] = None, encode=None
):
    """
    Converting a base64 to ndarray. For images use "b64_to_img"
    :param byte_array:
    :param dtype: pass None if dtype is encoded in the byte_array
    :param shape: pass None if shape is encoded in the byte_array
    :param logger: logger instance
    :param encode: whether encode param or not!
    :return:
    """
    import base64
    import struct

    if encode:
        byte_array = byte_array.encode(encode)
    elif not encode and isinstance(byte_array, str):
        byte_array = byte_array.encode()

    if dtype is None and byte_array[:6].decode("utf-8") == "dtype:":
        dtype_len = struct.unpack(">I", byte_array[6:10])[0]
        dtype = byte_array[10: dtype_len + 10].decode("utf-8")
        byte_array = byte_array[dtype_len + 10:]
    elif dtype is None:
        dtype = np.float32
        log_print(
            logger,
            f"dtype is not defined in byte_array nor as an input, setting dtype to {dtype}",
        )
    elif isinstance(dtype, str):
        pass
    elif isinstance(dtype, type):
        dtype = str(dtype)
    else:
        value_error_log(logger, f"dtype: {dtype} is not supported!")

    if shape is None and byte_array[:6].decode("utf-8") == "shape:":
        shape_len = struct.unpack(">I", byte_array[6:10])[0]
        shape = struct.unpack(f">{shape_len // 4}I",
                              byte_array[10: 10 + shape_len])
        byte_array = byte_array[shape_len + 10:]
    elif shape is None:
        log_print(logger, f"shape is not defined in byte_array nor as input")
    elif isinstance(shape, tuple) or isinstance(shape, list):
        pass
    else:
        value_error_log(logger, f"shape: {shape} is not supported!")
    if isinstance(byte_array, str):
        byte_array = bytes(byte_array, encoding="utf-8")
    ndarray = np.frombuffer(base64.decodebytes(byte_array), dtype=dtype)
    if shape is not None:
        ndarray = ndarray.reshape(shape)
    return ndarray



import os
import shutil
from abc import ABC
from os.path import join
from typing import Dict, List, Union

from tqdm import tqdm

from deep_utils.main_abs import MainClass
from deep_utils.utils.box_utils.boxes import Box
from deep_utils.utils.dict_named_tuple_utils import dictnamedtuple
from deep_utils.utils.dir_utils.dir_utils import (
    dir_train_test_split,
    remove_create,
    transfer_directory_items,
    file_incremental
)
from deep_utils.utils.logging_utils import log_print
from deep_utils.utils.opencv_utils.main import show_destroy_cv2
from deep_utils.utils.os_utils.os_path import split_extension
from deep_utils.utils.shutil_utils.shutil_utils import mv_or_copy

OUTPUT_CLASS = dictnamedtuple(
    "Object", ["class_indices", "boxes", "confidences", "class_names", "elapsed_time"]
)


class OutputType:
    class_indices: List[int]
    boxes: List[List[int]]
    confidences: List[float]
    class_names: List[str]
    elapsed_time: float


class YOLOObjectDetector(MainClass, ABC):
    """
    This class contains handy functions for yolo versions!
    """

    @staticmethod
    def test_label_dir(dataset_dir, rename_dict: Union[Dict[int, str], None], images_name="images",
                       labels_name="labels"):
        images_path = join(dataset_dir, images_name)
        for name in sorted(os.listdir(images_path)):
            img_address = join(images_path, name)
            text_address = join(dataset_dir, labels_name, split_extension(name, '.txt'))
            YOLOObjectDetector.test_label(img_address, text_address, rename_dict=rename_dict)

    @staticmethod
    def test_label(img_path, label_path, rename_dict: Union[Dict[int, str], None] = None, show=True):
        import cv2

        img = cv2.imread(img_path)
        boxes, texts, orgs = [], [], []
        with open(label_path, mode="r") as f:
            for line in f.readlines():
                label, xc, yc, w, h = line.strip().split()
                xc, yc, w, h = float(xc), float(yc), float(w), float(h)
                boxes.append([xc, yc, w, h])
                if rename_dict is not None:
                    label = rename_dict.get(int(label), label)
                texts.append(f"label: {label}")
        boxes = Box.box2box(boxes,
                            in_format=Box.BoxFormat.XCYC,
                            to_format=Box.BoxFormat.XYXY,
                            in_source="CV",
                            to_source="Numpy",
                            in_relative=True,
                            to_relative=False,
                            shape_source="Numpy",
                            shape=img.shape[:2],
                            )
        img = Box.put_box_text(img, boxes, texts)
        if show:
            show_destroy_cv2(img)
        return img

    @staticmethod
    def split_dataset(
            base_dir,
            out_dir,
            mode="cp",
            test_size=0.2,
            remove_out_dir=False,
            skip_transfer=True,
            remove_in_dir=False,
            image_based=True
    ):
        """

        :param base_dir:
        :param out_dir:
        :param mode:
        :param test_size:
        :param remove_out_dir:
        :param skip_transfer: If the file does not exist, skip and do not raise Error
        :param remove_in_dir: if mode is mv and this is set to true the in_dir will be removed
        :param image_based: If set to True, the split will be done based on images not labels
        :return:
        """
        split_identifier_name = "images" if image_based else "labels"
        split_identifier_name_prime = "labels" if image_based else "images"
        identifier_train_names, identifier_val_names = dir_train_test_split(
            join(base_dir, split_identifier_name),
            train_dir=join(out_dir, "train", split_identifier_name),
            val_dir=join(out_dir, "val", split_identifier_name),
            mode=mode,
            remove_out_dir=remove_out_dir,
            test_size=test_size,
            remove_in_dir=remove_in_dir
        )
        if image_based:
            identifier_train_prime = [os.path.splitext(name)[0] + ".txt" for name in identifier_train_names]
            identifier_val_prime = [os.path.splitext(name)[0] + ".txt" for name in identifier_val_names]
        else:
            img_names_dict = {split_extension(name)[0]: name for name in
                              os.listdir(join(base_dir, split_identifier_name_prime))}
            identifier_train_prime = [img_names_dict[split_extension(name)[0]] for name in identifier_train_names]
            identifier_val_prime = [img_names_dict[split_extension(name)[0]] for name in identifier_val_names]

        transfer_directory_items(
            join(base_dir, split_identifier_name_prime),
            join(out_dir, "train", split_identifier_name_prime),
            identifier_train_prime,
            mode=mode,
            remove_out_dir=remove_out_dir,
            skip_transfer=skip_transfer,
            remove_in_dir=False,  # by removing dir at this point, there would be no data for the next transfer :)
        )
        transfer_directory_items(
            join(base_dir, split_identifier_name_prime),
            join(out_dir, "val", split_identifier_name_prime),
            identifier_val_prime,
            mode=mode,
            remove_out_dir=remove_out_dir,
            skip_transfer=skip_transfer,
            remove_in_dir=remove_in_dir,
        )
        if remove_in_dir:
            shutil.rmtree(base_dir)

    @staticmethod
    def extract_label(label_path, img_path=None, shape=None, shape_source=None):
        with open(label_path, mode="r") as f:
            boxes, labels = [], []
            for line in f.readlines():
                label, b1, b2, b3, b4 = line.strip().split()
                boxes.append([float(b1), float(b2), float(b3), float(b4)])
                labels.append(int(label))

        if img_path is not None:
            import cv2
            shape = cv2.imread(img_path).shape[:2]
            shape_source = "Numpy"
        if shape is not None and shape_source is not None:
            boxes = Box.box2box(
                boxes,
                in_source=Box.BoxSource.CV,
                to_source="Numpy",
                in_format="XCYC",
                to_format="XYXY",
                in_relative=True,
                to_relative=False,
                shape=shape,
                shape_source=shape_source,
            )
        return boxes, labels

    @staticmethod
    def clean_samples(label_path, img_path, logger=None, verbose=1):
        """
        Using this method, one can remove images and the labels do not have their correspondences.
        :param label_path:
        :param img_path:
        :param logger:
        :param verbose:
        :return:
        """
        c = 0
        image_names, image_extensions = [], []
        for img in os.listdir(img_path):
            image_name, image_ext = os.path.splitext(img)
            image_names.append(image_name)
            image_extensions.append(image_ext)

        label_names = [os.path.splitext(l)[0] for l in os.listdir(label_path)]
        for label in label_names:
            if label not in image_names:
                remove_lbl_path = join(label_path, label + ".txt")
                log_print(logger, f"Removed {remove_lbl_path}", verbose=verbose)
                os.remove(remove_lbl_path)
                c += 1

        for img_name, img_ext in zip(image_names, image_extensions):
            if img_name not in label_names:
                remove_img_path = join(img_path, img_name + img_ext)
                log_print(logger, f"Removed {remove_img_path}", verbose=verbose)
                os.remove(remove_img_path)
                c += 1

        log_print(logger, f"Removed {c} samples!", verbose=verbose)

    @staticmethod
    def combine_datasets(
            dataset_paths: List[str],
            final_dataset_path: str,
            remove_final_dataset: bool = False,
            extra_index_format="prefix"
    ):
        """

        :param dataset_paths:
        :param final_dataset_path:
        :param remove_final_dataset:
        :param extra_index_format: whether to add suffix, prefix, or nothing as an extra index to data
        :return:
        """
        remove_create(final_dataset_path, remove_final_dataset)

        for dataset_path in dataset_paths:
            YOLOObjectDetector.transfer_dataset(dataset_path, final_dataset_path, extra_index_format)

    @staticmethod
    def transfer_dataset(dataset_path, final_dataset_path, extra_index_format="prefix", mode="cp", logger=None,
                         verbose=1):
        """
        This method is used to transfer samples from a dataset to other one
        :param dataset_path:
        :param final_dataset_path:
        :param extra_index_format:
        :param mode: Whether copy or move the dataset
        :param logger:
        :param verbose:
        :return:
        """
        images_path = os.path.join(dataset_path, "images")
        labels_path = os.path.join(dataset_path, "labels")

        final_images = os.path.join(final_dataset_path, "images")
        final_labels = os.path.join(final_dataset_path, "labels")

        os.makedirs(final_images, exist_ok=True)
        os.makedirs(final_labels, exist_ok=True)

        if not os.path.isdir(images_path) or not os.path.isdir(labels_path):
            log_print(logger, f"{images_path} or {labels_path} do not exit!", verbose=verbose)
            return

        # to make sure it's not replicated
        # index = len(os.listdir(final_images)) + len(os.listdir(final_labels))
        img_dict = dict()
        for img_name in tqdm(
                os.listdir(images_path),
                total=len(os.listdir(images_path)),
                desc=f"copying {images_path} to {final_images}",
        ):
            img_path = join(images_path, img_name)
            # new_name = split_extension(img_name, artifact_type=extra_index_format, artifact_value=img_index)
            new_path = join(final_images, img_name)
            new_path = file_incremental(new_path, artifact_type=extra_index_format)
            mv_or_copy(img_path, new_path, mode=mode)
            # key: original name without extension, val: new name without extension
            img_dict[split_extension(img_name)[0]] = split_extension(os.path.split(new_path)[-1])[0]

        for lbl_img_bare_name in tqdm(
                os.listdir(labels_path),
                total=len(os.listdir(labels_path)),
                desc=f"copying {labels_path} to {final_labels}"):
            lbl_path = os.path.join(labels_path, lbl_img_bare_name)
            lbl_bare_name = split_extension(lbl_img_bare_name)[0]
            lbl_img_bare_name = img_dict.get(lbl_bare_name, None)
            if lbl_img_bare_name is None:
                log_print(logger, f"label: {lbl_img_bare_name} has no image counterpart", verbose=verbose)
                continue
            mv_or_copy(lbl_path, os.path.join(final_labels, lbl_img_bare_name + ".txt"), mode=mode)
        log_print(logger, f"Successfully {mode}-ed {dataset_path} to {final_dataset_path}")

    @staticmethod
    def rename_labels(labels_dir, rename_dict: dict):
        """
        rename labels of a directory of labels based on the input rename dictionary
        :param labels_dir:
        :param rename_dict: A dictionary by which the input labels will be renamed
        :return:
        """
        assert os.path.isdir(labels_dir), "The input is not a directory"

        for label_name in os.listdir(labels_dir):
            if label_name.endswith(".txt"):
                label_path = join(labels_dir, label_name)
                with open(label_path, mode='r') as read_file:
                    labels = []
                    for line in read_file.readlines():
                        line = line.strip()
                        label, *box = line.split(" ")
                        labels.append(" ".join([str(rename_dict.get(int(label), label)), *box]))
                with open(label_path, mode='w') as write_file:
                    write_file.writelines(labels)



import torch
from deep_utils.utils.logging_utils.logging_utils import log_print
from deep_utils.utils.object_utils.object_utils import get_obj_variables


class TorchUtils:
    @staticmethod
    def save_config_to_weight(weight_path, config, logger=None, verbose=1):

        best_weight = torch.load(weight_path)
        for k, v in get_obj_variables(config).items():
            if k not in best_weight:
                best_weight[k] = v
            else:
                log_print(logger, f"[Warning] Did not save {k} = {v} because there is a variable with the same name!",
                          verbose=verbose)
        torch.save(best_weight, weight_path)



class Config:
    def __init__(self):
        self.RBF_url = (
            "https://github.com/Practical-AI/deep_utils/releases/download/0.2.0/RBF.zip"
        )
        self.RBF_cache = "weights/vision/face_detection/ultra-light/tf/RBF/RBF.zip"
        self.RBF = None
        self.slim_url = "https://github.com/Practical-AI/deep_utils/releases/download/0.2.0/slim.zip"
        self.slim_cache = "weights/vision/face_detection/ultra-light/tf/slim/slim.zip"
        self.slim = None



import sys


def subset_sum(
    input_numbers,
    target_number,
    result=(None, None, sys.maxsize),
    partial=None,
    indices=None,
    index=0,
):
    """
    Returns a subset of the inputs that has the closest value to the target_number input.
    Output: indices of chosen samples, values of chosen samples, difference from the input target_number.
    :param input_numbers:
    :param target_number:
    :param result:
    :param partial:
    :param indices:
    :param index:
    :return:
    >>> subset_sum([3, 9, 8, 4, 5, 7, 10], 150)
    [(0, 1, 2, 3, 4, 5, 6), (3, 9, 8, 4, 5, 7, 10), 104]
    >>> subset_sum([3, 9, 8, 4, 5, 7, 10], 15)
    [(0, 2, 3), (3, 8, 4), 0]
    >>> subset_sum([0, 0, 2], 1)
    [(2,), (2,), 1]
    """
    indices = [] if indices is None else indices
    partial = [] if partial is None else partial
    result = list(result) if not isinstance(result, list) else result
    result[0] = [] if result[0] is None else result[0]
    result[1] = [] if result[1] is None else result[1]

    s = sum(partial)

    # check if the partial sum is equals to target
    diff = abs(s - target_number)

    if diff == 0:
        result[0] = tuple(indices)
        result[1] = tuple(partial)
        result[-1] = abs(target_number - sum(partial))
        return
    elif diff <= result[-1]:
        result[0] = tuple(indices)
        result[1] = tuple(partial)
        result[-1] = abs(target_number - sum(partial))

    for i in range(len(input_numbers)):
        current_index = i + index
        n = input_numbers[i]
        remaining = input_numbers[i + 1:]
        subset_sum(
            remaining,
            target_number,
            result,
            partial + [n],
            indices + [current_index],
            current_index + 1,
        )
        if result[-1] == 0 or sum(partial) - target_number > result[-1]:
            break
    return result



import os
import shutil
from os.path import join
from pathlib import Path
from typing import Dict, List, Tuple, Union

from deep_utils.utils.shutil_utils.shutil_utils import mv_or_copy
from deep_utils.utils.logging_utils import log_print, value_error_log
from deep_utils.utils.os_utils.os_path import split_extension


def transfer_directory_items(
        in_dir,
        out_dir,
        transfer_list,
        mode="cp",
        remove_out_dir=False,
        skip_transfer=False,
        remove_in_dir=False,
):
    """

    Args:
        in_dir:
        out_dir:
        transfer_list:
        mode:
        remove_out_dir:
        skip_transfer: If the file does not exist, skip and do not raise Error.
        remove_in_dir: if mode is mv and this is set to true the in_dir will be removed!

    Returns:

    """
    print(f"[INFO] Starting to copying/moving from {in_dir} to {out_dir}")
    remove_create(out_dir, remove=remove_out_dir)
    if mode == "cp":
        for name in transfer_list:
            try:
                shutil.copy(os.path.join(in_dir, name), out_dir)
            except FileNotFoundError as e:
                if skip_transfer:
                    print("[INFO] shutil.copy did not find the file, skipping...")
                else:
                    raise FileNotFoundError(e)
    elif mode == "mv":
        for name in transfer_list:
            try:
                shutil.move(os.path.join(in_dir, name), out_dir)
            except FileNotFoundError as e:
                if skip_transfer:
                    print("[INFO] shutil.move did not find the file, skipping...")
                else:
                    raise FileNotFoundError()
        if remove_in_dir:
            shutil.rmtree(in_dir)
    else:
        raise ValueError(
            f"{mode} is not supported, supported modes: mv and cp")
    print(f"finished copying/moving from {in_dir} to {out_dir}")


def dir_train_test_split(
        in_dir,
        train_dir="./train",
        val_dir="./val",
        test_size=0.1,
        mode="cp",
        remove_out_dir=False,
        skip_transfer=False,
        remove_in_dir=False,
):
    """

    :param in_dir:
    :param train_dir:
    :param val_dir:
    :param test_size:
    :param mode:
    :param remove_out_dir:
    :param skip_transfer: If the file does not exist, skip and do not raise Error
    :param remove_in_dir: if mode is mv and this is set to true the in_dir will be removed!
    :return:
    """
    from sklearn.model_selection import train_test_split

    list_ = os.listdir(in_dir)
    train_name, val_name = train_test_split(list_, test_size=test_size)
    transfer_directory_items(
        in_dir,
        train_dir,
        train_name,
        mode=mode,
        remove_out_dir=remove_out_dir,
        skip_transfer=skip_transfer,
        remove_in_dir=False,
    )
    transfer_directory_items(
        in_dir,
        val_dir,
        val_name,
        mode=mode,
        remove_out_dir=remove_out_dir,
        skip_transfer=skip_transfer,
        remove_in_dir=remove_in_dir,
    )
    return train_name, val_name


def split_dir_of_dir(
        in_dir,
        train_dir="./train",
        val_dir="./val",
        test_size=0.1,
        mode="cp",
        remove_out_dir=False,
        remove_in_dir=False,
):
    """

    Args:
        in_dir:
        train_dir:
        val_dir:
        test_size:
        mode:
        remove_out_dir:
        remove_in_dir: if mode is mv and this is set to true the in_dir will be removed!

    Returns:

    """
    if remove_out_dir:
        remove_create(train_dir)
        remove_create(val_dir)
    for data in os.listdir(in_dir):
        dir_ = join(in_dir, data)
        if dir_ in [train_dir, val_dir]:
            print(
                f"[INFO] {dir_} is equal to {val_dir} or {train_dir}, Skipping ...")
            continue
        if not os.path.isdir(dir_):
            print(f"[INFO] {dir_} is not a directory, Skipping ...")
            continue
        if len(os.listdir(dir_)) == 0:
            print(f"[INFO] {dir_} is empty, Skipping ...")
            continue
        dir_train_test_split(
            dir_,
            train_dir=join(train_dir, data),
            val_dir=join(val_dir, data),
            mode=mode,
            test_size=test_size,
            remove_out_dir=remove_out_dir,
            remove_in_dir=remove_in_dir,
        )
    if mode == "mv" and remove_in_dir:
        shutil.rmtree(in_dir)


def split_xy_dir(
        x_in_dir,
        y_in_dir,
        x_train_dir="train/samples",
        y_train_dir="train/targets",
        x_val_dir="val/samples",
        y_val_dir="val/targets",
        mode="cp",
        val_size=0.1,
        skip_transfer=False,
        remove_out_dir=False,
        remove_in_dir=False,
):
    train_names, val_names = dir_train_test_split(
        x_in_dir,
        train_dir=x_train_dir,
        val_dir=x_val_dir,
        mode=mode,
        remove_out_dir=remove_out_dir,
        test_size=val_size,
    )
    train_labels = [os.path.splitext(name)[0] + ".txt" for name in train_names]
    val_labels = [os.path.splitext(name)[0] + ".txt" for name in val_names]

    transfer_directory_items(
        y_in_dir,
        y_train_dir,
        train_labels,
        mode=mode,
        remove_out_dir=remove_out_dir,
        skip_transfer=skip_transfer,
        remove_in_dir=remove_in_dir,
    )
    transfer_directory_items(
        y_in_dir,
        y_val_dir,
        val_labels,
        mode=mode,
        remove_out_dir=remove_out_dir,
        skip_transfer=skip_transfer,
        remove_in_dir=remove_in_dir,
    )


def crawl_directory_dataset(
        dir_: str,
        ext_filter: list = None,
        map_labels=False,
        label_map_dict: dict = None,
        logger=None,
        verbose=1,
) -> Union[Tuple[List[str], List[int]], Tuple[List[str], List[int], Dict]]:
    """
    crawls a directory of classes and returns the full path of the items paths and their class names
    :param dir_: path to directory of classes
    :param ext_filter: extensions that will be passed and others will be dropped
    :param map_labels: This map the labels to indices
    :param label_map_dict: A map which is used for filtering and also mapping the labels!
    :param logger: A logger
    :param verbose: whether print logs or not!
    :return: Tuple[List[str], List[int], Dict], x_list containing the paths, y_list containing the class_names, and
    label_map dictionary.
    """
    print(f"[INFO] beginning to crawl {dir_}")
    x, y = [], []
    label_map = dict()
    for cls_name in os.listdir(dir_):
        cls_path = join(dir_, cls_name)
        if not os.path.isdir(cls_path):
            continue
        for item_name in os.listdir(cls_path):
            item_path = join(cls_path, item_name)
            name, ext = os.path.splitext(item_name)
            if ext_filter is not None and ext not in ext_filter:
                log_print(
                    logger,
                    f"{item_path} with {ext} is not in ext_filtering: {ext_filter}",
                    verbose=verbose,
                )
                continue

            if label_map_dict is not None:
                if cls_name in label_map_dict:
                    y.append(label_map_dict[cls_name])
                else:
                    log_print(logger, f"Skipping cls_name:{cls_name}, x: {x}")
                    continue
            elif map_labels:
                if cls_name not in label_map:
                    cls_index = len(label_map)
                    label_map[cls_name] = cls_index
                else:
                    cls_index = label_map[cls_name]
                y.append(cls_index)
            else:
                y.append(cls_name)
            x.append(item_path)
    log_print(logger, f"successfully crawled {dir_}", verbose=verbose)
    if map_labels:
        return x, y, label_map
    else:
        return x, y


def remove_create(dir_, remove=True, logger=None, verbose=1):
    """
    Removes and creates the input directory!
    :param dir_:
    :param remove: whether to remove the directory or not
    :param logger:
    :param verbose:
    :return:
    """
    import os
    import shutil

    if os.path.exists(dir_) and remove:
        shutil.rmtree(dir_)
    if dir_:
        os.makedirs(dir_, exist_ok=True)
        log_print(logger, f"Successfully removed and created dir: {dir_}", verbose=verbose)


def mkdir_incremental(dir_path: str, base_name="exp", fix_name=None) -> Path:
    """
    makes new directories, if it exists increment it and makes another one. Good for hyperparameter tuning!
    Args:
        dir_path:
        base_name:
        fix_name: If provided this will be created!

    Returns:

    """
    os.makedirs(dir_path, exist_ok=True)
    if fix_name is not None:
        final_path = os.path.join(dir_path, fix_name)
        os.makedirs(final_path, exist_ok=True)
    else:
        folders = []
        for dir_ in os.listdir(dir_path):
            if base_name in dir_:
                counter = dir_.split(base_name + "_")[-1]
                if counter.isdigit():
                    folders.append(int(counter))
        if len(folders) == 0:
            final_path = os.path.join(dir_path, base_name + f"_1")
        else:
            max_counter = max(folders)
            final_path = os.path.join(
                dir_path, base_name + f"_{max_counter + 1}")
        os.makedirs(final_path)

    return Path(final_path)


def file_incremental(file_path, artifact_type="prefix", artifact_value=0, extra_punctuation="_",
                     add_artifact_value=False):
    """
    This function is used to increment a file's address with prefix or suffix values until it becomes unique
    :param file_path:
    :param artifact_type:
    :param artifact_value:
    :param extra_punctuation:
    :param add_artifact_value: If set to True, adds the artifact_value then checks its existence.
    :return:
    """
    dir_, n = os.path.split(file_path)
    artifact_value = int(artifact_value)
    while True:
        if add_artifact_value:
            # Maybe someone requires their file to have the first artifact
            file_path = join(dir_, split_extension(n, artifact_type=artifact_type, artifact_value=artifact_value,
                                                   extra_punctuation=extra_punctuation))
        if not os.path.exists(file_path):
            break

        if not add_artifact_value:
            file_path = join(dir_, split_extension(n, artifact_type=artifact_type, artifact_value=artifact_value,
                                                   extra_punctuation=extra_punctuation))
        artifact_value += 1
    return file_path


def cp_mv_all(
        input_dir,
        res_dir,
        mode="cp",
        filter_ext: Union[list, tuple, str, None] = None,
        artifact_type="prefix",
        artifact_value=0,
        extra_punctuation="_",
        add_artifact_value=False,
        logger=None,
        verbose=1,

):
    """
    mv/cp all the files in a directory to another one. In case any of the files had the same name as the target files,
     their name will be incremented using file_incremental
    :param input_dir:
    :param res_dir:
    :param mode:
    :param filter_ext:
    :return:
    """
    n = 0
    filter_ext = [filter_ext] if filter_ext is not None and isinstance(filter_ext, str) else filter_ext
    for f_name in os.listdir(input_dir):
        _, ext = split_extension(f_name)
        if filter_ext is not None and ext not in filter_ext:
            continue
        f_in_path = os.path.join(input_dir, f_name)
        f_out_path = os.path.join(res_dir, f_name)
        f_out_path = file_incremental(f_out_path, artifact_type=artifact_type, artifact_value=artifact_value,
                                      extra_punctuation=extra_punctuation, add_artifact_value=add_artifact_value)

        mv_or_copy(f_in_path, f_out_path, mode=mode)
        n += 1
    log_print(
        logger,
        f"Successfully {mode}-ed {n} items with filters: {filter_ext} from {input_dir} to {res_dir}",
        verbose=verbose,
    )


def split_segmentation_dirs(
        in_images,
        in_masks,
        out_train="./train",
        out_val="./val",
        image_dir_name="images",
        mask_dir_name="masks",
        img_ext=None,
        mask_ext=None,
        mode="cp",
        test_size=0.2,
        remove_out_dir=False,
        remove_in_dir=False,
        skip_transfer=False,
):
    from sklearn.model_selection import train_test_split

    if img_ext is None and mask_ext is None:
        in_image_list = os.listdir(in_images)
        in_mask_list = os.listdir(in_masks)
        in_mask_dict = {split_extension(i)[0]: i for i in in_mask_list}

        in_image_list_train, in_image_list_val = train_test_split(
            in_image_list, test_size=test_size
        )
        in_mask_list_train = [
            in_mask_dict[split_extension(i)[0]] for i in in_image_list_train
        ]
        in_mask_list_val = [
            in_mask_dict[split_extension(i)[0]] for i in in_image_list_val
        ]
    else:
        in_image_list = [i for i in os.listdir(
            in_images) if i.endswith(img_ext)]
        in_image_list_train, in_image_list_val = train_test_split(
            in_image_list, test_size=test_size
        )
        in_mask_list_train = [
            split_extension(i, extension=mask_ext) for i in in_image_list_train
        ]
        in_mask_list_val = [
            split_extension(i, extension=mask_ext) for i in in_image_list_val
        ]

    transfer_directory_items(
        in_images,
        join(out_train, image_dir_name),
        in_image_list_train,
        mode=mode,
        remove_out_dir=remove_out_dir,
        skip_transfer=skip_transfer,
        remove_in_dir=remove_in_dir,
    )
    transfer_directory_items(
        in_images,
        join(out_val, image_dir_name),
        in_image_list_val,
        mode=mode,
        remove_out_dir=remove_out_dir,
        skip_transfer=skip_transfer,
        remove_in_dir=remove_in_dir,
    )

    transfer_directory_items(
        in_masks,
        join(out_train, mask_dir_name),
        in_mask_list_train,
        mode=mode,
        remove_out_dir=remove_out_dir,
        skip_transfer=skip_transfer,
        remove_in_dir=remove_in_dir,
    )
    transfer_directory_items(
        in_masks,
        join(out_val, mask_dir_name),
        in_mask_list_val,
        mode=mode,
        remove_out_dir=remove_out_dir,
        skip_transfer=skip_transfer,
        remove_in_dir=remove_in_dir,
    )


def find_file(dir_path, name, ext=".ckpt", logger=None, verbose=1):
    """
    finds the closest file to the input name in the input dir_path
    :param dir_path:
    :param name:
    :param ext:
    :param logger:
    :param verbose:
    :return:
    """
    for file_name in os.listdir(dir_path):
        if file_name.startswith(name) and file_name.endswith(ext):
            log_print(
                logger, f"name: {file_name} found in dir: {dir_path}", verbose=verbose
            )
            return join(dir_path, file_name)
    log_print(
        logger,
        f"name: {name} not found in dir: {dir_path}",
        log_type="warning",
        verbose=verbose,
    )
    return None


def combine_directory_of_directories(dataset_dir, result_dir, remove_result_dir=True, logger=None, verbose=0):
    """
    Combine a directory of directories in which there are several samples, and they should be combined in a final
    directory
    :param dataset_dir:
    :param result_dir:
    :param remove_result_dir:
    :param logger:
    :param verbose:
    :return:
    """
    remove_create(result_dir, remove_result_dir)
    for directory_name in os.listdir(dataset_dir):
        directory = join(dataset_dir, directory_name)
        if os.path.isdir(directory):
            os.system(f'cp {directory}/* {result_dir}')
            log_print(logger, f'Successfully moved {directory} to {result_dir}!', verbose=verbose)



import numpy as np

from deep_utils.utils.box_utils.boxes import Point


def resize(img, dsize, in_source="Numpy", mode="cv2", interpolation=None):
    mode = "cv2" if mode is None else mode
    if mode == "cv2":
        dsize = Point.point2point(
            dsize, in_source=in_source, to_source=Point.PointSource.CV
        )
        new_img = cv2_resize(img, dsize=dsize, interpolation=interpolation)
    elif mode.lower() == "pil":
        from PIL import Image

        img = img if isinstance(img, Image.Image) else Image.fromarray(img)
        new_img = img.resize(dsize, interpolation)
    else:
        raise ValueError(f"{mode} is not supported, supported types: cv2, ")
    return new_img


def cv2_resize(img, dsize, dst=None, fx=None, fy=None, interpolation=None):
    import cv2

    if len(img.shape) == 3:
        return cv2.resize(img, dsize, dst, fx, fy, interpolation)
    elif len(img.shape) == 4:
        return np.array(
            [cv2.resize(im, dsize, dst, fx, fy, interpolation) for im in img]
        )


def get_img_shape(img):
    if len(img.shape) == 4:
        return img.shape
    elif len(img.shape) == 3:
        return np.expand_dims(img, axis=0).shape
    else:
        raise Exception(f"shape: {img.shape} is not an image")


def resize_ratio(
    img,
    size: int,
    pad: bool = False,
    mode: str = "cv2",
    pad_val: int = 0,
    return_pad: bool = False,
):
    """
    Resize an image while keeping aspect-ratio
    Args:
        img: input image
        size: max-size
        mode: cv2 or pil
        pad: Pad the smaller size to become equal to the bigger one.
        pad_val: Value for padding
        return_pad: returns pad values for further processes, default is false. return: image, (h_top, h_bottom, w_left, w_right)

    Returns: resized image

    """
    if mode == "cv2":
        import math

        import cv2

        h, w = img.shape[:2]
        ratio = h / w
        if ratio > 1:
            h = size
            w = int(size / ratio)
            h_top, h_bottom = 0, 0
            w_pad = size - w
            w_left, w_right = math.ceil(w_pad / 2), w_pad // 2
        elif ratio < 1:
            h = int(size * ratio)
            w = size
            w_left, w_right = 0, 0
            h_pad = size - h
            h_top, h_bottom = math.ceil(h_pad / 2), h_pad // 2
        else:
            h, w = size, size
            w_left, w_right, h_bottom, h_top = 0, 0, 0, 0
        image = cv2.resize(img, (w, h), interpolation=cv2.INTER_NEAREST)
        if pad:
            image = cv2.copyMakeBorder(
                image,
                h_top,
                h_bottom,
                w_left,
                w_right,
                cv2.BORDER_CONSTANT,
                value=pad_val,
            )
            if return_pad:
                return image, (h_top, h_bottom, w_left, w_right)
        return image
    else:
        raise ValueError(f"Requested mode: {mode} is not supported!")


def resize_save_aspect_ratio(img, size):
    import torch.nn.functional as F
    from PIL import Image
    from torchvision import transforms

    w, h = img.size
    ratio = h / w
    if ratio > 1:
        h = size
        w = int(size / ratio)
    elif ratio < 1:
        h = int(size * ratio)
        w = size
    else:
        return img.resize((size, size), Image.NEAREST)

    image = img.resize((w, h), Image.NEAREST)
    im = transforms.ToTensor()(image)
    val = h - w
    if val < 0:
        a = 0
        b = int(-val / 2)
    else:
        a = int(val / 2)
        b = 0

    result = F.pad(input=im, pad=(a, a, b, b))
    image = transforms.ToPILImage()(result)

    return image


if __name__ == "__main__":
    import cv2

    img = cv2.imread(
        "/home/ai/projects/Irancel-Service/hard_samples/0013903470.jpg")
    print(f"[INFO] input shape: {img.shape}")
    img = resize_ratio(img, 900)
    print(f"[INFO] output shape: {img.shape}")
    cv2.imshow("", img)
    cv2.waitKey(0)



import sys
import time
from pathlib import Path
from typing import Dict, List, Type, Union

import numpy as np
import torch

from deep_utils.utils.box_utils.boxes import Box
from deep_utils.utils.dict_named_tuple_utils import dictnamedtuple
from deep_utils.utils.lib_utils.lib_decorators import (
    in_shape_fix,
    lib_rgb2bgr,
    out_shape_fix,
)
from deep_utils.vision.object_detection.yolo.yolo_detector import YOLOObjectDetector
from .config import Config

FILE = Path(__file__).resolve()
ROOT = FILE.parents[0]  # YOLOv5 root directory
if str(ROOT) not in sys.path:
    sys.path.append(str(ROOT))  # add ROOT to PATH
from .models.experimental import attempt_load
from .utils_.datasets import letterbox
from .utils_.general import non_max_suppression, scale_boxes

OUTPUT_CLASS = dictnamedtuple(
    "Object", ["class_indices", "boxes", "confidences", "class_names", "elapsed_time"]
)


class OutputType:
    class_indices: List[int]
    boxes: List[List[int]]
    confidences: List[float]
    class_names: List[str]
    elapsed_time: float


class YOLOV5TorchObjectDetector(YOLOObjectDetector):
    def __init__(
            self,
            class_names=None,
            model_weight=None,
            device="cpu",
            img_size=(320, 320),
            confidence=0.4,
            iou_thresh=0.45,
            **kwargs,
    ):
        super(YOLOV5TorchObjectDetector, self).__init__(
            name=self.__class__.__name__,
            file_path=__file__,
            class_names=class_names,
            model_weight=model_weight,
            device=device,
            img_size=img_size,
            confidence=confidence,
            iou_thresh=iou_thresh,
            **kwargs,
        )
        self.config: Config

    @staticmethod
    def yolo_resize(
            img,
            new_shape=(640, 640),
            color=(114, 114, 114),
            auto=True,
            scaleFill=False,
            scaleup=True,
    ):

        return letterbox(
            img,
            new_shape=new_shape,
            color=color,
            auto=auto,
            scaleFill=scaleFill,
            scaleup=scaleup,
        )

    def load_model(self):
        # if self.config.model_weight:
        self.model = attempt_load(
            self.config.model_weight, device=self.config.device
        )
        self.model.to(self.config.device)
        self.model.eval()
        img = torch.zeros((1, 3, *self.config.img_size), device=self.config.device)
        self.model(img)
        print(f"{self.name}: weights are loaded")

    def detect_objects(
            self,
            img,
            is_rgb,
            class_indices=None,
            confidence=None,
            iou_thresh=None,
            img_size=None,
            agnostic=None,
            get_time=False,
    ) -> Union[Type[OutputType], Dict[str, list]]:
        """

        :param img:
        :param is_rgb: Is used with rgb2bgr. The required conversion is done automatically.
        :param confidence:
        :param iou_thresh:
        :param class_indices: target class indices, the rest will be ignored!
        :param agnostic:
        :param get_time:
        :param img_size:
        :param kwargs:
        :return:
        """

        tic = time.time() if get_time else 0

        self.update_config(
            confidence=confidence, iou_thresh=iou_thresh, img_size=img_size
        )
        img = lib_rgb2bgr(img, target_type="rgb", is_rgb=is_rgb)
        img = in_shape_fix(img, size=4)

        im0 = img
        img = np.array(
            [self.yolo_resize(im, new_shape=self.config.img_size)[0] for im in im0]
        )
        img = img.transpose((0, 3, 1, 2))
        img = np.ascontiguousarray(img)
        img = torch.from_numpy(img).to(self.config.device)
        img = img / 255.0
        with torch.no_grad():
            prediction = self.model(img, augment=False)[0]
        prediction = non_max_suppression(
            prediction,
            self.config.confidence,
            self.config.iou_thresh,
            classes=class_indices,
            agnostic=agnostic,
        )
        boxes, class_names, classes, confidences = [
            [[] for _ in range(im0.shape[0])] for _ in range(4)
        ]
        for i, det in enumerate(prediction):  # detections per image
            if len(det):
                det[:, :4] = scale_boxes(
                    img.shape[2:], det[:, :4], im0[i].shape
                ).round()
                for *xyxy, conf, cls in reversed(det):
                    bbox = Box.box2box(
                        xyxy,
                        in_source=Box.BoxSource.Torch,
                        to_source=Box.BoxSource.Numpy,
                        return_int=True,
                    )
                    boxes[i].append(bbox)
                    confidences[i].append(round(conf.item(), 2))
                    cls = int(cls.item())
                    classes[i].append(cls)
                    class_names[i].append(self.config.class_names[cls])

        if get_time:
            toc = time.time()
            elapsed_time = toc - tic
        else:
            elapsed_time = 0

        output = OUTPUT_CLASS(
            class_indices=classes,
            boxes=boxes,
            confidences=confidences,
            class_names=class_names,
            elapsed_time=elapsed_time,
        )
        output = out_shape_fix(output)
        return output


from typing import List


def bio_tagger(tokens: List[str], label, others_label="o"):
    if label == others_label:
        return ["o"] * len(tokens)
    return [f"B-{label}"] + [f"I-{label}"] * (len(tokens) - 1)



def get_hash_file(file, buffer_size=65536):
    """
    This function converts a file in chunk mode to hash file!
    buffer_size = 65536 lets read stuff in 64kb chunks
    :param file: str, io.BufferedReader
    :param buffer_size:
    :return:
    """
    import hashlib
    hash_obj = hashlib.sha512()
    if isinstance(file, str):
        with open(file, 'rb') as f:
            while True:
                data = f.read(buffer_size)
                if not data:
                    break
                hash_obj.update(data)
    else:
        hash_obj = hashlib.sha512()
        while True:
            data = file.read(buffer_size)
            if not data:
                break
            hash_obj.update(data)
    return hash_obj.hexdigest()



import pytest

from deep_utils import repeat_dimension


@pytest.mark.basic
def test_repeat_dimension():
    """
    Testing the repeat dimension function
    Returns: Nothing

    """
    import numpy as np

    dummy_arrays = [
        np.random.randint(0, 255, (1200, 900, 1), dtype=np.uint8),
        np.random.randint(0, 255, (800, 1200), dtype=np.uint8),
    ]

    preferred_outputs = [
        (1200, 900, 3),
        (800, 1200, 3),
    ]

    for dummy_array, preferred_output in zip(dummy_arrays, preferred_outputs):
        out = repeat_dimension(dummy_array, n=3, d=2)
        assert (
            out.shape == preferred_output
        ), f"repeat_dimension failed for input array.shape={dummy_array.shape} returned: {out.shape}"

    # failures
    with pytest.raises(Exception) as e:
        # size failure
        repeat_dimension(np.random.randint(0, 255, (800,)), d=2)
        # type failure
        repeat_dimension(10, d=2)



import torch.nn as nn
import math
import torch.nn.functional as F

__all__ = ['SENet']


def conv3x3(in_planes, out_planes, stride=1):
    """3x3 convolution with padding"""
    return nn.Conv2d(in_planes, out_planes, kernel_size=3, stride=stride,
                     padding=1, bias=False)


# This SEModule is not used.
class SEModule(nn.Module):

    def __init__(self, planes, compress_rate):
        super(SEModule, self).__init__()
        self.conv1 = nn.Conv2d(planes, planes // compress_rate, kernel_size=1, stride=1, bias=True)
        self.conv2 = nn.Conv2d(planes // compress_rate, planes, kernel_size=1, stride=1, bias=True)
        self.relu = nn.ReLU(inplace=True)
        self.sigmoid = nn.Sigmoid()

    def forward(self, x):
        module_input = x
        x = F.avg_pool2d(module_input, kernel_size=module_input.size(2))
        x = self.conv1(x)
        x = self.relu(x)
        x = self.conv2(x)
        x = self.sigmoid(x)
        return module_input * x


class BasicBlock(nn.Module):
    expansion = 1

    def __init__(self, inplanes, planes, stride=1, downsample=None):
        super(BasicBlock, self).__init__()
        self.conv1 = conv3x3(inplanes, planes, stride)
        self.bn1 = nn.BatchNorm2d(planes)
        self.relu = nn.ReLU(inplace=True)
        self.conv2 = conv3x3(planes, planes)
        self.bn2 = nn.BatchNorm2d(planes)
        self.downsample = downsample
        self.stride = stride

    def forward(self, x):
        residual = x

        out = self.conv1(x)
        out = self.bn1(out)
        out = self.relu(out)

        out = self.conv2(out)
        out = self.bn2(out)

        if self.downsample is not None:
            residual = self.downsample(x)

        out += residual
        out = self.relu(out)

        return out


class Bottleneck(nn.Module):
    expansion = 4

    def __init__(self, inplanes, planes, stride=1, downsample=None):
        super(Bottleneck, self).__init__()
        self.conv1 = nn.Conv2d(inplanes, planes, kernel_size=1, stride=stride, bias=False)
        self.bn1 = nn.BatchNorm2d(planes)
        self.conv2 = nn.Conv2d(planes, planes, kernel_size=3, stride=1, padding=1, bias=False)
        self.bn2 = nn.BatchNorm2d(planes)
        self.conv3 = nn.Conv2d(planes, planes * 4, kernel_size=1, bias=False)
        self.bn3 = nn.BatchNorm2d(planes * 4)
        self.relu = nn.ReLU(inplace=True)
        self.downsample = downsample
        self.stride = stride

        # SENet
        compress_rate = 16
        # self.se_block = SEModule(planes * 4, compress_rate)  # this is not used.
        self.conv4 = nn.Conv2d(planes * 4, planes * 4 // compress_rate, kernel_size=1, stride=1, bias=True)
        self.conv5 = nn.Conv2d(planes * 4 // compress_rate, planes * 4, kernel_size=1, stride=1, bias=True)
        self.sigmoid = nn.Sigmoid()

    def forward(self, x):
        residual = x

        out = self.conv1(x)
        out = self.bn1(out)
        out = self.relu(out)

        out = self.conv2(out)
        out = self.bn2(out)
        out = self.relu(out)

        out = self.conv3(out)
        out = self.bn3(out)

        ## senet
        out2 = F.avg_pool2d(out, kernel_size=out.size(2))
        out2 = self.conv4(out2)
        out2 = self.relu(out2)
        out2 = self.conv5(out2)
        out2 = self.sigmoid(out2)
        # out2 = self.se_block.forward(out)  # not used

        if self.downsample is not None:
            residual = self.downsample(x)

        out = out2 * out + residual
        # out = out2 + residual  # not used
        out = self.relu(out)
        return out


class SENet(nn.Module):

    def __init__(self, block=Bottleneck, layers=(3, 4, 6, 3), num_classes=8631, include_top=True):
        self.inplanes = 64
        super(SENet, self).__init__()
        self.include_top = include_top

        self.conv1 = nn.Conv2d(3, 64, kernel_size=7, stride=2, padding=3, bias=False)
        self.bn1 = nn.BatchNorm2d(64)
        self.relu = nn.ReLU(inplace=True)
        self.maxpool = nn.MaxPool2d(kernel_size=3, stride=2, padding=0, ceil_mode=True)

        self.layer1 = self._make_layer(block, 64, layers[0])
        self.layer2 = self._make_layer(block, 128, layers[1], stride=2)
        self.layer3 = self._make_layer(block, 256, layers[2], stride=2)
        self.layer4 = self._make_layer(block, 512, layers[3], stride=2)
        self.avgpool = nn.AvgPool2d(7, stride=1)
        self.fc = nn.Linear(512 * block.expansion, num_classes)

        for m in self.modules():
            if isinstance(m, nn.Conv2d):
                n = m.kernel_size[0] * m.kernel_size[1] * m.out_channels
                m.weight.data.normal_(0, math.sqrt(2. / n))
            elif isinstance(m, nn.BatchNorm2d):
                m.weight.data.fill_(1)
                m.bias.data.zero_()

    def _make_layer(self, block, planes, blocks, stride=1):
        downsample = None
        if stride != 1 or self.inplanes != planes * block.expansion:
            downsample = nn.Sequential(
                nn.Conv2d(self.inplanes, planes * block.expansion,
                          kernel_size=1, stride=stride, bias=False),
                nn.BatchNorm2d(planes * block.expansion),
            )

        layers = []
        layers.append(block(self.inplanes, planes, stride, downsample))
        self.inplanes = planes * block.expansion
        for i in range(1, blocks):
            layers.append(block(self.inplanes, planes))

        return nn.Sequential(*layers)

    def forward(self, x):
        x = self.conv1(x)
        x = self.bn1(x)
        x = self.relu(x)
        x = self.maxpool(x)
        x = self.layer1(x)
        x = self.layer2(x)
        x = self.layer3(x)
        x = self.layer4(x)

        x = self.avgpool(x)

        if not self.include_top:
            return x

        x = x.view(x.size(0), -1)
        x = self.fc(x)
        return x


import tensorflow as tf


def basic_conv(
    x,
    out_ch,
    kernel_size,
    stride=(1, 1),
    padding=0,
    dilation=1,
    relu=True,
    bn=True,
    prefix="basic_conv",
):
    if 0 < padding:
        out = tf.keras.layers.ZeroPadding2D(padding=padding, name=f"{prefix}_padding")(
            x
        )
    else:
        out = x
    out = tf.keras.layers.Conv2D(
        out_ch,
        kernel_size,
        strides=stride,
        dilation_rate=dilation,
        use_bias=(not bn),
        name=f"{prefix}_conv",
    )(out)
    if bn:
        out = tf.keras.layers.BatchNormalization(
            epsilon=1e-5, name=f"{prefix}_bn")(out)
    if relu:
        out = tf.keras.layers.ReLU(name=f"{prefix}_relu")(out)

    return out


def basic_rfb(
    x, in_ch, out_ch, stride=1, scale=0.1, map_reduce=8, vision=1, prefix="basic_rfb"
):
    inter_ch = in_ch // map_reduce

    branch0 = basic_conv(
        x, inter_ch, kernel_size=1, stride=1, relu=False, prefix=f"{prefix}.branch0.0"
    )
    branch0 = basic_conv(
        branch0,
        2 * inter_ch,
        kernel_size=3,
        stride=stride,
        padding=1,
        prefix=f"{prefix}.branch0.1",
    )
    branch0 = basic_conv(
        branch0,
        2 * inter_ch,
        kernel_size=3,
        stride=1,
        dilation=vision + 1,
        padding=vision + 1,
        relu=False,
        prefix=f"{prefix}.branch0.2",
    )

    branch1 = basic_conv(
        x, inter_ch, kernel_size=1, stride=1, relu=False, prefix=f"{prefix}.branch1.0"
    )
    branch1 = basic_conv(
        branch1,
        2 * inter_ch,
        kernel_size=3,
        stride=stride,
        padding=1,
        prefix=f"{prefix}.branch1.1",
    )
    branch1 = basic_conv(
        branch1,
        2 * inter_ch,
        kernel_size=3,
        stride=1,
        dilation=vision + 2,
        padding=vision + 2,
        relu=False,
        prefix=f"{prefix}.branch1.2",
    )

    branch2 = basic_conv(
        x, inter_ch, kernel_size=1, stride=1, relu=False, prefix=f"{prefix}.branch2.0"
    )
    branch2 = basic_conv(
        branch2,
        (inter_ch // 2) * 3,
        kernel_size=3,
        stride=1,
        padding=1,
        prefix=f"{prefix}.branch2.1",
    )
    branch2 = basic_conv(
        branch2,
        2 * inter_ch,
        kernel_size=3,
        stride=stride,
        padding=1,
        prefix=f"{prefix}.branch2.2",
    )
    branch2 = basic_conv(
        branch2,
        2 * inter_ch,
        kernel_size=3,
        stride=1,
        dilation=vision + 4,
        padding=vision + 4,
        relu=False,
        prefix=f"{prefix}.branch2.3",
    )

    out = tf.keras.layers.Concatenate(axis=-1, name=f"{prefix}_cat")(
        [branch0, branch1, branch2]
    )
    out = basic_conv(
        out, out_ch, kernel_size=1, stride=1, relu=False, prefix=f"{prefix}.convlinear"
    )
    shortcut = basic_conv(
        x, out_ch, kernel_size=1, stride=stride, relu=False, prefix=f"{prefix}.shortcut"
    )
    out = tf.multiply(out, scale, name=f"{prefix}_mul")
    out = tf.keras.layers.Add(name=f"{prefix}_add")([out, shortcut])
    out = tf.keras.layers.ReLU(name=f"{prefix}_relu")(out)

    return out


def separable_conv(x, out_ch, kernel_size, stride, padding, prefix="separable_conv"):
    out = tf.keras.layers.ZeroPadding2D(
        padding=padding, name=f"{prefix}_dconv_padding"
    )(x)

    out = tf.keras.layers.DepthwiseConv2D(
        kernel_size, strides=stride, name=f"{prefix}_dconvbias"
    )(out)
    out = tf.keras.layers.ReLU(name=f"{prefix}_relu")(out)
    out = tf.keras.layers.Conv2D(out_ch, 1, name=f"{prefix}_convbias")(out)

    return out


def conv_bn(x, out_ch, stride, padding=1, prefix="conv_bn"):
    out = tf.keras.layers.ZeroPadding2D(
        padding=padding, name=f"{prefix}.0_padding")(x)
    out = tf.keras.layers.Conv2D(
        out_ch, (3, 3), strides=stride, use_bias=False, name=f"{prefix}.0_conv"
    )(out)
    out = tf.keras.layers.BatchNormalization(
        epsilon=1e-5, name=f"{prefix}.1_bn")(out)
    out = tf.keras.layers.ReLU(name=f"{prefix}.2_relu")(out)

    return out


def conv_dw(x, out_ch, stride, padding=1, prefix="conv_dw"):
    out = tf.keras.layers.ZeroPadding2D(
        padding=padding, name=f"{prefix}.0_padding")(x)
    out = tf.keras.layers.DepthwiseConv2D(
        3, strides=stride, use_bias=False, name=f"{prefix}.0_dconv"
    )(out)
    out = tf.keras.layers.BatchNormalization(
        epsilon=1e-5, name=f"{prefix}.1_bn")(out)
    out = tf.keras.layers.ReLU(name=f"{prefix}.2_relu")(out)

    out = tf.keras.layers.Conv2D(out_ch, 1, use_bias=False, name=f"{prefix}.3_conv")(
        out
    )
    out = tf.keras.layers.BatchNormalization(
        epsilon=1e-5, name=f"{prefix}.4_bn")(out)
    out = tf.keras.layers.ReLU(name=f"{prefix}.5_relu")(out)

    return out



import numpy as np

from deep_utils.utils.logging_utils import log_print, value_error_log


def repeat_dimension(input_array: np.ndarray, n=3, d=2, logger=None, verbose=0):
    """

    Args:
        input_array:
        n: how many times to repeat
        d: dimension to be repeated!
        logger:
        verbose: whether print or not

    Returns:

    """
    if not isinstance(input_array, np.ndarray):
        value_error_log(logger, "input_array is not of type array")
    elif len(input_array.shape) == d:
        input_array = np.expand_dims(input_array, axis=-1)
    elif len(input_array.shape) < d:
        value_error_log(
            logger,
            f"input_array.shape:{input_array.shape} is lower the one:{d} which should be repeated!",
        )
    result = np.repeat(input_array, n, axis=d)
    log_print(
        logger, f"Successfully create result.shape: {result.shape}!", verbose=verbose
    )
    return result



from typing import TypeVar, Union

from sqlalchemy import select

SQLAlchemyUtilsType = TypeVar("SQLAlchemyUtilsType", bound="SQLAlchemyUtils")


class SQLAlchemyUtils:

    @staticmethod
    def select_id(cls, id_, session, class_id_feature="id") -> Union[SQLAlchemyUtilsType, None]:
        try:
            selected = select(cls).where(getattr(cls, class_id_feature) == id_)
            instance = session.scalars(selected).one()
        except:
            return None
        return instance

    @staticmethod
    def update(instance, session):
        try:
            session.add(instance)
            session.commit()
        except:
            session.rollback()
        return instance



from ._version import __version__
from .audio import *
from .augmentation import *
from .blocks import *
from .callbacks import *
from .design_patterns import *
from .gis import *
from .nlp import *
from .vision import *
from .elasticsearch import *
from .utils import *



from __future__ import annotations

from abc import ABC, abstractmethod
from typing import Any, List, Tuple, Union
from deep_utils.utils.logging_utils.logging_utils import log_print


class CoRError(Exception):
    def __init__(
            self, message: Union[str, None] = None, code: Union[int, None] = 0, output: Union[None, dict] = None
    ):
        """
        This is a custom error used to break the responsibility chain that is running through the manager!
        It helps to increase codes' readability; when one sees this error realizes that the process is broke because
        of an unfinished process that's handled but not in the desired and intended format
        :param message: message send by the user, default is None
        :param code: code send by the user, default is None
        :param output: A dictionary that contains output.
        """
        if output is None:
            output = None
        super().__init__(message)
        self.message = message
        self.code = code
        self.output = output


class Handler(ABC):
    """
    The Handler interface declares a method for building the chain of handlers.
    It also declares a method for executing a request.
    """

    @abstractmethod
    def set_next(self, handler: Handler) -> Handler:
        pass

    @abstractmethod
    def handle(self, request) -> dict:
        pass


class CoRHandler(Handler):
    """
    The default chaining behavior can be implemented inside a base handler
    class.
    """

    _next_handler: Handler = None

    def set_next(self, handler: Handler) -> Handler:
        """

        :param handler:
        :return:
        """
        self._next_handler = handler
        # Returning a handler from here will let us link handlers in a
        return handler

    @abstractmethod
    def handle(self, request: Any) -> dict:
        if self._next_handler:
            return self._next_handler.handle(request)

        raise NotImplementedError(
            "next_handler is missing, and the module is not implemented!"
        )


class CoRManger(CoRHandler):
    def __init__(
            self,
            handlers: Union[Tuple[CoRHandler], List[CoRHandler]],
            internal_error_code=-1,
    ):
        """
        The default manger, which takes in a list of handlers and process it
        :param handlers: input handlers
        :param internal_error_code: This value is returned when handlers raise an unexpected error. Default of -1
        is used for internal errors, don't use it as a code in the handlers to prevent confusion.
        """
        assert handlers, "handlers can not be None or empty"
        next_handler: Union[CoRHandler, None] = None
        handler = None
        for handler in handlers[::-1]:
            assert isinstance(
                handler, CoRHandler
            ), "Handlers should have CoRHandler type"
            if next_handler is not None:
                handler.set_next(next_handler)
            next_handler = handler

        self.handler: CoRHandler = handler
        self.internal_error_code = internal_error_code

    def handle(self, request: Any, logger=None, verbose=1):
        try:
            output = self.handler.handle(request)
        except CoRError as e:
            output = dict(e.output)
            if e.code:
                output["code"] = e.code
            if e.message:
                output["message"] = e.message
        except BaseException as e:
            output = dict()
            output["code"] = self.internal_error_code
            output["message"] = f"Internal Error --> type: {type(e)}, msg: {e}"
        log_print(
            logger,
            f"Request code: {output.get('code', 0)}, message: {output.get('message', '')}",
            verbose=verbose,
        )
        return output



from typing import Union


def get_request(ip, down_key, down_message="Down") -> dict:
    """
    This simple function wraps the get request
    :param ip:
    :param down_key: key of the output if it fails
    :param down_message:
    :return:
    """
    import requests
    try:
        status = requests.get(ip).json()
    except:
        status = {down_key: down_message}
    return status


async def post_json(input_url, data: dict, header="application/json"):
    """
    Async format for sending requests
    :param input_url:
    :param data:
    :param header:
    :return:
    """
    import httpx
    async with httpx.AsyncClient() as client:
        response = await client.post(input_url, json=data,
                                     headers={"Content-Type": header})

        return response.json()


def _get_file_content(file_path: str, file_content_type: str = 'multipart/form-data') -> tuple:
    import os
    file_name = os.path.basename(file_path)
    file_content = open(file_path, 'rb')
    return file_name, file_content, file_content_type


async def post_form(input_url, data_key: str, data_path: str):
    """
    Async format for sending form requests!
    :param input_url:
    :param data_path:
    :param data_key:
    :return:
    """
    import httpx
    file_name, file_content, file_content_type = _get_file_content(data_path)
    async with httpx.AsyncClient() as client:
        response = await client.post(input_url, files={data_key: (file_name, file_content, file_content_type)})
        return response.json()


async def post_form_upload(input_url, data_key: str, upload_file):
    """
    Async format for sending form requests. It gets upload file as input
    :param input_url:
    :param upload_file:
    :param data_key:
    :return:
    """
    import httpx
    async with httpx.AsyncClient() as client:
        response = await client.post(input_url, files={
            data_key: (upload_file.filename, upload_file.file, upload_file.content_type)})
        return response.json()



import os
from typing import Union

IMG_EXTENSIONS = [
    ".jpg",
    ".JPG",
    ".jpeg",
    ".JPEG",
    ".png",
    ".PNG",
    ".ppm",
    ".PPM",
    ".bmp",
    ".BMP",
    ".tiff",
]


def is_image_file(filename):
    return any(filename.endswith(extension) for extension in IMG_EXTENSIONS)


def split_extension(path, extension: Union[str, None] = None,
                    suffix: Union[str, None] = None,
                    prefix: Union[str, None] = None,
                    artifact_type: Union[str, None] = None,
                    artifact_value: Union[str, int, None] = None,
                    extra_punctuation: Union[str, None] = None
                    ):
    """
    split extension or add new suffix, prefix, or extension
    :param path:
    :param extension:
    :param suffix:
    :param prefix:
    :param artifact_type: prefix or suffix
    :param artifact_value: value for defined prefix or suffix
    :param extra_punctuation: the punctuation to be used before suffix or after prefix
    :return:

    >>> split_extension("image.jpg", suffix="_res")
    'image_res.jpg'
    >>> split_extension("image.jpg")
    ('image', '.jpg')
    >>> split_extension("image.jpg", extension=".png")
    'image.png'
    >>> split_extension("image.jpg", extension="png")
    'image.png'
    >>> split_extension("image.jpg", extension="png", prefix="0_")
    '0_image.png'
    >>> split_extension("image.jpg", extension="png", suffix="_res", prefix="0_")
    '0_image_res.png'
    """
    artifact_value = str(artifact_value) if artifact_value is not None else None
    assert (artifact_value and artifact_type) or (
            not artifact_value and not artifact_type), "both artifact type and artifact value should be None or not None"

    if (artifact_value and artifact_type) and (suffix or prefix):
        raise ValueError("artifact value and artifact type cannot have values while suffix and prefix are not None")

    extra_punctuation = extra_punctuation if extra_punctuation else ""
    if artifact_type == "suffix":
        return _split_extension(path, extension=extension, suffix=extra_punctuation + artifact_value)
    elif artifact_type == "prefix":
        return _split_extension(path, extension=extension, prefix=artifact_value + extra_punctuation)
    elif artifact_type is None:
        return _split_extension(path, extension=extension, suffix=extra_punctuation + suffix if suffix else None,
                                prefix=prefix + extra_punctuation if prefix else None)
    else:
        raise ValueError(f"input artifact_type:{artifact_type} is invalid!")


def _split_extension(path, extension: Union[str, None] = None,
                     suffix: Union[str, None] = None,
                     prefix: Union[str, None] = None):
    """
    split extension or add new suffix, prefix, or extension
    :param path:
    :param extension:
    :param suffix:
    :param prefix:
    :return:

    >>> split_extension("image.jpg", suffix="_res")
    'image_res.jpg'
    >>> split_extension("image.jpg")
    ('image', '.jpg')
    >>> split_extension("image.jpg", extension=".png")
    'image.png'
    >>> split_extension("image.jpg", extension="png")
    'image.png'
    >>> split_extension("image.jpg", extension="png", prefix="0_")
    '0_image.png'
    >>> split_extension("image.jpg", extension="png", suffix="_res", prefix="0_")
    '0_image_res.png'
    """
    remain, extension_ = os.path.splitext(path)
    prefix = prefix if prefix else ""
    suffix = suffix if suffix else ""
    extension = extension if extension else ""

    extension = '.' + extension if extension and not extension.startswith(".") else extension
    core = prefix + remain + suffix

    if extension:
        return core + extension

    if not suffix and not prefix:
        return core, extension_
    else:
        return core + extension_


def split_all(path):
    all_parts = []
    while 1:
        parts = os.path.split(path)
        if parts[0] == path:  # sentinel for absolute paths
            all_parts.insert(0, parts[0])
            break
        elif parts[1] == path:  # sentinel for relative paths
            all_parts.insert(0, parts[1])
            break
        else:
            path = parts[0]
            all_parts.insert(0, parts[1])
    return all_parts


def get_file_name(file):
    """
    Get the file's name
    :param file: the file in which code is running. Path in `__file__`
    :return:
    """
    return os.path.splitext(os.path.split(file)[-1])[0]



from deep_utils.main_abs.dummy_framework.dummy_framework import DummyObject, requires_backends


class HaarcascadeCV2FaceDetector(metaclass=DummyObject):
    _backend = ["cv2"]

    def __init__(self, *args, **kwargs):
        requires_backends(self, self._backend)


class MTCNNTFFaceDetector(metaclass=DummyObject):
    _backend = ["tensorflow", "cv2"]

    def __init__(self, *args, **kwargs):
        requires_backends(self, self._backend)


class MTCNNTorchFaceDetector(metaclass=DummyObject):
    _backend = ["torch", "cv2"]

    def __init__(self, *args, **kwargs):
        requires_backends(self, self._backend)


class RetinaFaceTorchFaceDetector(metaclass=DummyObject):
    _backend = ["torch", "cv2"]

    def __init__(self, *args, **kwargs):
        requires_backends(self, self._backend)


class SSDCV2CaffeFaceDetector(metaclass=DummyObject):
    _backend = ["cv2"]

    def __init__(self, *args, **kwargs):
        requires_backends(self, self._backend)


class UltralightTFFaceDetector(metaclass=DummyObject):
    _backend = ["tensorflow", "cv2"]

    def __init__(self, *args, **kwargs):
        requires_backends(self, self._backend)


class UltralightTorchFaceDetector(metaclass=DummyObject):
    _backend = ["torch", "cv2"]

    def __init__(self, *args, **kwargs):
        requires_backends(self, self._backend)



try:
    from deep_utils.dummy_objects.audio.diarization.pyannote import PyannoteAudioDiarization
    from .pyannote_audio_diarization import PyannoteAudioDiarization
except ModuleNotFoundError:
    pass



try:
    from deep_utils.dummy_objects.vision.face_detection import SSDCV2CaffeFaceDetector
    from .cv2.caffe.ssd_cv2_caffe_face_detection import SSDCV2CaffeFaceDetector
except ModuleNotFoundError:
    pass
